import os
import pytest

# pylint: disable=missing-docstring

from . import ConfigFixtures, StackFixtures
from ..config import Config


@pytest.mark.parametrize("config", ["empty"], indirect=True)
class TestEmptyConfig(ConfigFixtures):
    def test_empty(self, config):
        assert "environment" in config.vars
        assert "name" in config.vars

        assert not list(config.params.keys())


@pytest.mark.parametrize("config", ["simple"], indirect=True)
class TestSimpleConfig(ConfigFixtures):
    def test_simple_load_vars(self, config):
        assert set(config.vars.keys()) >= set(["foo", "bar"])
        assert config.var("bar") == "hello, world!"

    def test_simple_load_params(self, config):
        assert list(config.params.keys()) == ["jane"]
        assert config.param("jane") == 'Jane said "hello, world!"'


@pytest.mark.parametrize("config", ["includes"], indirect=True)
class TestIncludedConfig(ConfigFixtures):
    def test_include_vars(self, config):
        v = config.vars
        assert set(v.keys()) >= set(["a", "b", "c", "d"])

        assert v["a"] == "this is top-level"
        assert v["b"] == "this is from first.yaml"
        assert v["c"] == "this is from second.yaml"
        assert v["d"] == "this is from third.yaml"

    def test_include_params(self, config):
        v = config.params
        assert sorted(v.keys()) == ["p1", "p2", "p3"]

        assert v["p1"] == "this is top-level param"
        assert v["p2"] == "this is param from third.yaml"
        assert v["p3"] == 'interpolated "this is top-level"'


@pytest.mark.parametrize("config", ["environments-simple"], indirect=True)
class TestEnvironmentalPrecdenceConfig(ConfigFixtures):
    def test_var_precedence(self, config):
        v = config.vars
        assert v["a"] == "this is top-level environment - P1"
        assert v["b"] == "this is top-level default - P2"
        assert v["c"] == "this is included environment - P3"
        assert v["d"] == "this is included default - P4"

    def test_param_precedence(self, config):
        v = config.params
        assert v["p1"] == "this is top-level environment param - P1"
        assert v["p2"] == "this is top-level default param - P2"
        assert v["p3"] == "this is included environment param - P3"
        assert v["p4"] == "this is included default param - P4"


@pytest.mark.parametrize("config", ["templates"], indirect=True)
class TestEnvironmentalPrecdenceConfig(ConfigFixtures):
    def test_precedence(self, config):
        source = config.template_source

        assert source.name == "a_template"
        assert source.version == "main"

    def test_shortcut_usage(self, config):
        config = Config("shortcut", environment="test", config_path=self.fixture_path("config", "templates"))

        source = config.template_source

        assert source.name == "b_template"
        assert source.version == "main"


@pytest.mark.parametrize("config", ["tags"], indirect=True)
class TestTags(ConfigFixtures):
    def test_inheritance(self, config):
        tags = config.tags.to_list()

        assert {"Key": "Environment", "Value": "test"} in tags
        assert {"Key": "Product", "Value": "tags-test"} in tags
        assert {"Key": "StackTopLevel", "Value": "some-application"} in tags
        assert {"Key": "StackEnvSpecific", "Value": "True"} in tags


@pytest.mark.parametrize("config", ["core"], indirect=True)
class TestEnvironmentalPrecedenceConfig(ConfigFixtures):
    def test_stack_name(self, config):
        assert config.core.stack_name == "forced-stack-name"

    def test_define_valid_environments(self, config):
        assert config.core.environments == ["dev", "test", "prod"]

    def test_only_valid_environments_allowed(self, config):
        with pytest.raises(Exception) as ex:
            Config("invalid-env", environment="stage", config_path=self.fixture_path("config", "core"))

        assert "invalid-env.yaml defines environments 'stage', which are is listed in core.environments (dev, test, prod)"


@pytest.mark.parametrize("config", ["types"], indirect=True)
class TestInterpolatedDict(ConfigFixtures):
    def test_str(self, config):
        assert config.vars["a_str"] == "hello!"

    def test_int(self, config):
        assert config.vars["an_int"] == 12

    def test_false_preserved(self, config):
        assert config.vars["a_false"] == False

    def test_true_preserved(self, config):
        assert config.vars["a_true"] == True

    def test_none_preserved(self, config):
        assert config.vars["a_none"] == None
        assert config.vars["no_value"] == None


@pytest.mark.parametrize("config", ["simple"], indirect=True)
class TestSimpleConfig(ConfigFixtures):
    def test_simple_load_vars(self, config):
        assert set(config.vars.keys()) >= set(["foo", "bar"])
        assert config.var("bar") == "hello, world!"

    def test_simple_load_params(self, config):
        assert list(config.params.keys()) == ["jane"]
        assert config.param("jane") == 'Jane said "hello, world!"'


@pytest.mark.parametrize("config", ["simple"], indirect=True)
class TestSimpleConfig(ConfigFixtures):
    def test_aws_vars(self, config, aws, sts):
        assert config.vars["account_id"] == 123456789012



import codecs
import uuid
import io


_boundary = "//"

_content_types = {
    # Maps string prefix to mimetype
    "#!": "text/x-shellscript",
    "#cloud-boothook": "text/cloud-boothook",
    "#cloud-config": "text/cloud-config",
    "#cloud-config-archive": "text/cloud-config-archive",
    "#include": "text/x-include-url",
    "#include-once": "text/x-include-once-url",
    "#part-handler": "text/part-handler",
}


def multipart_encode(files: dict) -> str:
    """
    Given a dict of { filename => content }, returns byte array of UTF-8
    mime-encoded payload.
    """
    out = io.StringIO()

    out.write(f'Content-Type: multipart/mixed; boundary="{_boundary}"\n')
    out.write("MIME-Version: 1.0\n\n")

    for filename, content in files:
        content_type = guess_type(content.splitlines()[0])
        out.writelines(
            [
                "--" + _boundary + "\n",
                "MIME-Version: 1.0\n",
                "Content-Type: " + content_type + '; charset="utf-8"\n',
                'Content-Disposition: attachment; filename="' + filename + '"\n',
                "Content-Transfer-Encoding: utf-8\n",
                "\n",
                *content.splitlines(keepends=True),
                "\n\n",
            ]
        )
    out.write(f"--{_boundary}--\n")

    return out.getvalue()


def guess_type(first_line):
    for prefix, content_type in _content_types.items():
        if first_line.startswith(prefix):
            return content_type
    return "application/octet-stream"



from __future__ import annotations

import json
import zipfile

from io import BytesIO
from pytest import fixture

from . import StackFixtures
from ..config import Config
from ..provider import GenericProvider
from ..template import RenderedTemplate, TemplateWithConfig
from ..template_helpers import TemplateHelpers
from ..template_source import TemplateSource
from ..cfn_bucket import CfnBucket


class TestTemplateHelpers(StackFixtures):
    @fixture
    def env(self):
        class FakeEnvironment:
            def __init__(self):
                self.globals = dict()
                self.filters = dict()

        return FakeEnvironment()

    @fixture
    def config(self, sts):
        return Config("main", environment="test", config_path=self.fixture_path("custom_helpers", "config"))

    @fixture
    def provider(self):
        p = self.fixture_path("custom_helpers", "templates")
        return TemplateSource(name="main.yaml", root=p).provider()

    def test_custom_helpers_in_config(self, config):
        assert set(config.helpers) == set(["a_custom_helper", "this_one_should_appear_only_once", "another_custom_helper"])

    @fixture
    def basic_helpers(self, provider: GenericProvider, config: Config, env) -> TemplateHelpers:
        return TemplateHelpers(provider=provider, bucket=None, custom_helpers=[], config=config)

    def test_custom_helpers_loaded_from_provider(self, provider: GenericProvider, config: Config, env):
        helpers = TemplateHelpers(provider=provider, bucket=None, custom_helpers=["a_custom_helper"], config=config)

        helpers.inject_helpers(env)

        assert "a_custom_helper" in env.globals
        assert env.globals["a_custom_helper"](41) == 42

    def test_custom_helpers_available_in_template(self, sts, provider: GenericProvider, config: Config, env):
        template = TemplateWithConfig(provider=provider, config=config)

        rendered = template.render()

        assert type(rendered) == RenderedTemplate

        assert rendered["foo_should_be_42"] == "42"
        assert rendered["resourcify"] == "FooBarBaz"

    def test_lambda_uri(self, cfn_bucket, provider, config):
        bucket = CfnBucket(config.aws)
        helpers = TemplateHelpers(provider, bucket=bucket, custom_helpers=[], config=config)

        uri = helpers.lambda_uri("a_function")
        assert uri.startswith("s3://foo/functions/a_function/6fa3f1707cb555571039137d7970816f.zip")

        uri2 = helpers.lambda_uri("a_function")
        assert uri == uri2, "Generated URLs are deterministic"

    def test_upload_zip(self, cfn_bucket, provider, config):
        bucket = CfnBucket(config.aws)
        helpers = TemplateHelpers(provider, bucket=bucket, custom_helpers=[], config=config)

        # Upload our test directory. In contains a single file test.txt
        uri = helpers.upload_zip("files/test", prefix="/opt/foo")

        # Check it has correct url
        assert uri == "files/test/15315fc306560b9eb8332fe277bf8e95.zip"

        # Download file from S3 and check the file is valid ZIP and it contains the
        # expected content
        s3 = config.aws.resource("s3")
        object = s3.Object(cfn_bucket, "files/test/15315fc306560b9eb8332fe277bf8e95.zip")

        content = object.get()["Body"].read()
        fh = BytesIO(content)

        assert "/opt/foo/test.txt" in zipfile.ZipFile(fh, "r").namelist()

    def test_user_data(self, provider, config):
        helpers = TemplateHelpers(provider, bucket=None, custom_helpers=[], config=config)

        parsed = json.loads(helpers.user_data("test1"))

        assert "Fn::Base64" in parsed.keys()

        lines = parsed["Fn::Base64"]["Fn::Join"][1]
        assert "#!/bin/bash\n" in lines
        assert 'Content-Type: text/x-shellscript; charset="utf-8"\n' in lines
        assert {"Ref": "SomeResource"} in lines

    def test_package_ignore(self, provider, config):
        helpers = TemplateHelpers(provider, bucket=None, custom_helpers=[], config=config)

        # The template directory has some .package-ignore files that should be loaded
        ignore = helpers.ignore_list("functions/a_function")
        assert not ignore("foo")
        assert ignore("a")
        assert not ignore("b")
        assert ignore("c")
        assert not ignore("d")

        assert ignore(".package-ignore")



from pytest import fixture

from . import ConfigFixtures
from ..template_source import TemplateSource
from ..template import Template, RenderedTemplate
from ..provider import provider
from ..config import Config


class TestStackRefs(ConfigFixtures):
    @fixture
    def config(self, sts):
        return Config("main", environment="test", config_path=self.fixture_path("deploy-metadata", "config"))

    @fixture
    def provider(self, request):
        source = TemplateSource(name="main", root=self.fixture_path("deploy-metadata", "templates"))
        return provider(source)

    def test_basic_info(self, provider, config):
        t = Template(name="main", provider=provider, helpers=None)
        rendered = t.render(vars=config.vars)
        assert type(rendered) == RenderedTemplate

        assert rendered["Metadata"]["stack"]["deployed_at"].startswith("20")



from __future__ import annotations
from typing import Dict, Union

from botocore.exceptions import ClientError

from . import log
from .aws_config import AwsSettings
from .cfn_bucket import CfnBucket

class BasicStack:
    def __init__(self, aws: AwsSettings, name: str):
        self.aws = aws
        self.name = name
        self.cfn = aws.client("cloudformation")
        self.bucket = CfnBucket(aws)

    def exists(self):
        """Tests if stack with .name exists"""
        return self.status() != None

    def status(self):
        """Returns stack status, or None if stack does not exist"""
        stack = self.describe_stack()
        if stack:
            return stack["StackStatus"]

    def outputs(self) -> Union[Dict[str, BasicStack.Output], None]:
        """Return all outputs for a stack"""
        if not hasattr(self, "__outputs"):
            self.__outputs = self.__get_outputs()
        return self.__outputs

    def output(self, key) -> str:
        """Return value of named key, raising KeyError if no matching output found"""
        log.info("Getting output %s from %s", key, self.name)
        outputs = self.outputs()

        if key in outputs:
            return outputs[key]
        raise KeyError(f"{key} not in outputs of {self.name} - only have {list(outputs.keys())}")

    def describe_stack(self):
        """Perform a describe_stack operation against stack, returning None if it doesn't exist"""
        try:
            stack = self.cfn.describe_stacks(StackName=self.name)["Stacks"][0]
            assert self.name in [stack["StackId"], stack["StackName"]]
            return stack
        except ClientError as exc:
            err = exc.response["Error"]
            if (err["Code"] == "ValidationError") and ("does not exist" in err["Message"]):
                return None
            raise exc

    class Output(str):
        """
        A stack output can have a description (and other attributes). We are normally interested in just the
        value when querying the output, but the description is also available via an attribute

        output = stack.output('Foo')
        print("output description is " + output.description)

        """
        def __new__(cls, value: str, description: str):
            obj = str.__new__(cls, value)
            obj.description = description
            return obj

    def __get_outputs(self) -> Union[Dict[str, BasicStack.Output], None]:
        stack = self.describe_stack()
        if not stack:
            return None

        if "Outputs" not in stack:
            return None

        ret_val = {}
        for output in stack["Outputs"]:
            description = ""
            if "Description" in output:
                description = output["Description"]
            ret_val[output["OutputKey"]] = self.Output(output["OutputValue"], description)
        return ret_val



from dataclasses import dataclass

import boto3

from . import log


@dataclass
class AwsSettings:
    """Dataobject for AWS settings"""
    region: str
    cfn_bucket: str
    account_id: str = None
    profile: str = None
    role_arn: str = None
    web_identity_token: str = None
    aws_access_key: str = None
    aws_secret_key: str = None
    aws_session_token: str = None

    def __post_init__(self):
        self.__session = None
        self.__checked_account = False

    def client(self, service):
        """Returns boto client for given service"""
        session = self._session()
        log.info("client(%s), account_id=%s", service, self.account_id)
        return session.client(service, region_name=self.region)

    def resource(self, service):
        """Returns boto resource for given service"""
        session = self._session()
        log.info("resource(%s), account_id=%s", service, self.account_id)
        return session.resource(service, region_name=self.region)

    def get_account_id(self):
        """
        Ensure account-id has been retrieved
        """
        self._session()
        return self.account_id

    def _session(self):
        if not self.__session:
            session = self._session_from_access_key() or self._session_from_role() or self._session_from_profile() or self._session_from_env()

            if not self.__checked_account:
                sts = session.client("sts")
                account_id: str = sts.get_caller_identity()["Account"]
                if self.account_id:
                    if account_id != str(self.account_id):
                        raise ValueError(f"Incorrect AWS Account - expected {self.account_id}, but appear to be using {account_id} ")

                self.__checked_account = True
                self.account_id = account_id

            self.__session = session

        return self.__session

    def _session_from_role(self):
        """Assume role if explicitly set"""
        if self.role_arn and self.role_arn != "None":
            if self.web_identity_token and self.web_identity_token != "None":
                short_token = self.web_identity_token[:8] + '\u2026' * (len(self.web_identity_token) > 8)
                log.info("assuming role %s using web token %s", self.role_arn, short_token)
                role = boto3.client('sts').assume_role_with_web_identity(RoleArn=self.role_arn, RoleSessionName='stk-web-session-role', WebIdentityToken=self.web_identity_token)
            else:
                log.info("assuming role %s", self.role_arn)
                role = boto3.client('sts').assume_role(RoleArn=self.role_arn, RoleSessionName="stk-session-role")

            credentials = role['Credentials']
            return boto3.Session(credentials['AccessKeyId'], credentials['SecretAccessKey'], credentials['SessionToken'])

    def _session_from_profile(self):
        """Assume role if profile is set"""
        if self.profile:
            return boto3.Session(profile_name=str(self.profile))

    def _session_from_access_key(self):
        """Returns a boto3.Session if access keys are set"""
        access_keys = self._access_keys()
        if access_keys['aws_access_key_id'] and access_keys['aws_secret_access_key']:
            log.info("using sts access key")
            return boto3.Session(**access_keys)

    def _session_from_env(self):
        return boto3.Session()

    def _access_keys(self):
        """Returns a dict of access keys for use with boto3.Session, with 'None' (string) values set to None"""
        sts_auths = [self.aws_access_key, self.aws_secret_key, self.aws_session_token]

        result = {}
        for i, key in enumerate(["aws_access_key_id", "aws_secret_access_key", "aws_session_token"]):
            value = sts_auths[i]
            result[key] = value if value and value != "None" else None

        return result



"""
Base class for commands that work against a deployed/deployable stack
"""
import json
from typing import List
from dataclasses import dataclass
from rich.table import Table

import yaml
from . import console
from .stack import Stack
from .config import Config
from .util import parse_overrides

@dataclass
class StackDelegatedCommand:
    """
    Base class for commands that work against a deployed/deployable stack
    """
    name: str
    environment: str
    config_path: str
    template_path: str
    var: List
    param: List
    overrides: str
    outputs_format: str = "table"

    def __post_init__(self):
        overrides = parse_overrides(self.var, self.param, self.overrides)
        self.config = Config(
            name=self.name,
            environment=self.environment,
            config_path=self.config_path,
            template_path=self.template_path,
            overrides=overrides
        )
        self.stack = Stack(aws=self.config.aws,
                           name=self.config.core.stack_name)
        self.stack_name = self.stack.name

    def __getattr__(self, name):
        if hasattr(self.stack, name):
            return getattr(self.stack, name)
        return super().__getattr__(name)

    def show_outputs(self):
        """
        Display a table of outputs for a CFN stack
        """
        stack_outputs = self.outputs()
        if stack_outputs:
            if self.outputs_format == 'table':
                t = Table("Key", "Value", "Description", title="Stack Outputs",
                        title_justify="left", title_style="bold")
                for key in sorted(stack_outputs.keys()):
                    value = stack_outputs[key]
                    t.add_row(key, value, value.description)
                console.print(t)
            elif self.outputs_format == 'json':
                print(json.dumps(stack_outputs))
            elif self.outputs_format == 'yaml':
                print(yaml.dump({ k: str(v) for k, v in stack_outputs.items() }))
            else:
                raise ValueError(f"invalid output format {self.outputs_format}")
        else:
            console.print(f"Stack {self.stack_name} does not have any outputs")



from dataclasses import dataclass
from typing import Any, Dict
from jinja2 import Environment, StrictUndefined
from yaml import safe_load

class InterpolatedDict(dict):
    def __init__(self, object: Dict[str, Any], vars: Dict[str, Any]):
        # Handle loading from empty YAML file (results in None), or a 'config group' (e.g. params)
        # not being present - which is okay.
        if not object:
            return

        if type(object) != dict:
            raise Exception(object)

        env = Environment(undefined=StrictUndefined, extensions=["jinja2_strcase.StrcaseExtension"])

        for key, value in object.items():
            try:
                if value == None:
                    self[key] = None
                else:
                    value = env.from_string(str(value)).render(vars)
                    parsed_value = safe_load(value)
                    if parsed_value != None:
                        self[key] = parsed_value
            except Exception as ex:
                raise Exception(f"Unable to process {key}, value={object[key]}") from ex



@dataclass
class InterpolationError:
    """Captures information about an error occurring during evaluation of variables"""

    key: str
    value: str
    error: str




import difflib
import os
import pathlib
import sys
from typing import Any, Dict
from dataclasses import dataclass

import yaml
from rich.prompt import Prompt

from .. import clog, console
from ..provider import GitProvider


@dataclass
class RemoteFile:
    """
    Tuple of <content,provider>
    """
    content: str
    provider: GitProvider

    def template_config(self) -> dict:
        """
        returns config for template location that can be put into a config file; this exists
        due to inconsistent naming between internal fields and user config. Needs to be fixed.
        """
        return { "root": self.provider.root or "/", "version": self.provider.git_ref, "repo": self.provider.git_url }

class AddTemplateCmd:
    """
    Fetch information about template from remote repository and add configuration
    for it to this project.

    * Lots of assumptions, should work for simple cases
    """

    def __init__(self, repo_url: str, config_dir: str = ".") -> None:
        #
        self.config_dir = config_dir

        self.providers = []

        default = self._common("template")
        self.providers.append(
            GitProvider(
                    "upstream",
                    git_url=default["repo"],
                    root=default["root"],
                    git_ref=default["version"]
            )
        )

        if repo_url:
            self.providers.insert(
                0,
                GitProvider(
                    name="templates",
                    git_url=repo_url,
                    root="/",
                    git_ref="main"
                )
            )

        provider_urls = [provider.git_url for provider in self.providers]
        clog(f"Using { ', '.join(provider_urls) } for remote template source")

    def add(self, name: str, follow_refs: bool = True, inline: bool = False, local_template_dir: str = "templates") -> None:
        """
        Generate local configuration for remote template

        follow_refs: if true, will attempt to add referenced stacks (if they don't exist)
        """
        template_filename = name + ".yaml"

        # Load remote template
        try:
            template = self.remote_content(template_filename)
        except FileNotFoundError:
            # Refs use underscore (since yaml doesn't like hyphens in keys), but have been using hyphens
            # in stack names. Bit of a hack here :-/
            kebab_name = name.replace('_', '-')
            if kebab_name != name:
                clog(f"{template_filename}.yaml not found in remote repository, trying kebab-case version")
                self.add(kebab_name, follow_refs=follow_refs, inline=inline, local_template_dir=local_template_dir)
                return
            clog(f"{template_filename} not found in remote repository. Aborting")
            exit(-1)

        # Load meta-data associated with remote template
        #  file in template repository with _ prefix
        metadata_filename = "_" + template_filename

        metadata: Dict[str, Any] = {"config": {}}
        try:
            # Try to find metadata - only in the same provider, don't search self.providers
            template_metadata = self.remote_content(metadata_filename, provider=template.provider)
            metadata = dict(yaml.safe_load(template_metadata.content))
            # generate configuration file first (to avoid potential looks with auto-add)
            if metadata["config"] is None:
                metadata["config"] = {}
        except FileNotFoundError:
            clog(f"{metadata_filename} does not exist in remote repository - guessing defaults")

        # Build config object for this stack
        config = {
            "includes": ["common"],
            "environments": {x: None for x in self.core_environments()},
            **metadata.get("config"),  # type: ignore
        }

        if inline:
            # Make a local copy of the template under `local_template_dir`, and config file
            # will reference this local copy. Not ideal in the long term, but easy way to
            # get started
            clog(f"Writing local template to {os.path.join(local_template_dir, template_filename)}")
            config["template"] = { "root": local_template_dir, "version": None, "repo": None }
            self.write_local_file(local_template_dir, template_filename, content=template.content)
        else:
            # If template comes from repo that is not the default, override it in the config
            if template.template_config() != self._common()["template"]:
                config["template"] = template.template_config()

        # Wrote configuration file
        clog(f"Writing config file {name}.yml")
        content = [f"# Starting configuration for deploying stack {name}\n"]
        if "description" in metadata and metadata["description"]:
            description = ["# " + line for line in metadata["description"].splitlines(keepends=True)]
            content += description  # type: ignore

        content += [
            yaml.safe_dump(config, default_flow_style=False, explicit_start=True, sort_keys=False)
        ]

        self.write_local_file(name + ".yml", content="".join(content))

        # try to resolve referenced stacks. E.g. if "ec2" stack depends on "vpc", then we
        # automatically do an "add" for "vpc" stack.
        if follow_refs and "refs" in metadata["config"]:
            for ref in metadata["config"]["refs"].keys():
                if self.is_local_file(ref + ".yml") or self.is_local_file(ref + ".yaml"):
                    clog(f"skipping {ref} - config exists already")
                else:
                    self.add(ref, local_template_dir=local_template_dir)


    def is_local_file(self, *p) -> bool:
        """
        Returns true if path exists in config directory
        """
        local_file = os.path.join(self.config_dir, *p)
        return os.path.isfile(local_file)

    def remote_content(self, *p, provider=None) -> RemoteFile:
        """
        Returns content of remote file
        """
        if provider:
            if provider.is_file(*p):
                content = str(provider.content(*p), encoding="utf-8")
                return RemoteFile(content, provider)
        else:
            for provider in self.providers:
                if provider.is_file(*p):
                    content = str(provider.content(*p), encoding="utf-8")
                    return RemoteFile(content, provider)

        raise FileNotFoundError(f"unable to find {os.path.join(*p)} in any remotes")

    def local_content(self, *p) -> str:
        """
        Create local file (only if it doesn't exist)
        """
        local_file = os.path.join(self.config_dir, *p)
        with open(local_file, "r", encoding="utf-8") as fh:
            return fh.read()

    def write_local_file(self, *p, content: str):
        """
        Create local file (only if it doesn't exist)
        """
        local_file = os.path.join(self.config_dir, *p)
        pathlib.Path(local_file).parents[0].mkdir(parents=True, exist_ok=True)
        try:
            with open(local_file, "x", encoding="utf-8") as out:
                out.write(content)
            return
        except FileExistsError:
            with open(local_file, "r", encoding="utf-8") as inp:
                original_content = inp.read()
            if original_content == content:
                clog(f"...skipping {local_file}, file already exists and is identical")
                return

        # Print diff of current/proposed and prompt user if they want to overwrite
        diff = difflib.unified_diff(
            original_content.splitlines(keepends=True),
            content.splitlines(keepends=True),
            fromfile='current',
            tofile='proposed',
            lineterm="\n"
        )
        sys.stdout.writelines(diff)

        # Confirm with user whether they want to overwrite, skip, or abort
        answer = Prompt.ask(f"Overwrite {local_file} ?", console=console, choices=["y", "n", "q"], default="n")
        if answer == "y":
            with open(local_file, "w", encoding="utf-8") as out:
                out.write(content)
            return
        elif answer == "n":
            clog(f"skipping {local_file}")
        else:
            clog("Aborting")
            sys.exit(-1)

    def core_environments(self):
        """
        Return configured environments. By convention, these should be
        defined in includes/common.yaml
        """
        return self._common()["core"]["environments"]

    def _common(self, *keys):
        config = yaml.safe_load(self.local_content("includes/common.yml"))
        for key in keys:
            config = config.get(key, {})
        return config


if __name__ == "__main__":
    AddTemplateCmd("git@github.com:jwoffindin/stk-templates.git",
                   config_dir="./tmp").add("vpc")



from pytest import fixture, mark

from . import Fixtures
from ..template import RenderedTemplate, Template, FailedTemplate
from ..template_source import TemplateSource
from ..provider import provider


class TestTemplate(Fixtures):
    @fixture
    def provider(self, request):
        p = self.fixture_path("templates", "simple")
        source = TemplateSource(name=request.param, root=p)
        return provider(source)

    @mark.parametrize("provider", ["main.yaml"], indirect=True)
    def test_render_simple_template(self, provider):
        """
        The Template().render method returns a RenderedTemplate. The rendered template may
        have
        """
        t = Template(name="test", provider=provider, helpers=None)

        rendered = t.render(vars={"environment": "test", "result": "okay"})

        assert type(rendered) == RenderedTemplate

        # Rendered template acts like a dict for parsed template (assuming successfully parsed)
        assert rendered["template_version"] == 1.0
        assert rendered["interpolated"] == "Environment test is okay"

        assert "should_be_excluded" not in rendered

        assert "item_0" in rendered
        assert "item_1" in rendered
        assert "item_2" in rendered

        # str() method returns parsed template
        assert "interpolated: 'Environment test is okay'" in str(rendered)

    @mark.parametrize("provider", ["main.yaml"], indirect=True)
    def test_render_jinja_error_handling(self, provider):
        t = Template(provider=provider, custom_helpers=[])

        rendered = t.render(vars={})

        assert type(rendered) == FailedTemplate
        assert rendered.error
        assert "template_version: 1.0" in rendered.source

        # Not overly happy about overriding the 'str()' method to render an error
        as_str = str(rendered)

        assert "'environment' is undefined" in as_str
        assert "main.yaml at line 3" in as_str

    @mark.parametrize("provider", ["invalid.yaml"], indirect=True)
    def test_render_jinja_error_handling(self, provider):
        t = Template(name="test_template", provider=provider)

        rendered = t.render(vars={"environment": "test"})

        assert type(rendered) == FailedTemplate
        assert "Error occurred outsite of template" in str(rendered)

    @mark.parametrize("provider", ["quoted.yaml"], indirect=True)
    def test_render_quoted_template(self, provider):
        t = Template(name="quoted", provider=provider, helpers=None)

        rendered = t.render(vars={})

        assert "value: ['hello', 'there']" in str(rendered)



import yaml
from typing import List, Dict, Union

from . import log
from .config_file import ConfigDocument, ConfigFiles

def parse_overrides(cfg_vars: List[str], cfg_params: List[str], overrides: str) -> ConfigFiles:
    """
    Build a ConfigFiles list that provides ConfigFiles list from
    vars/params and generic 'overrides' dict.
    """
    log.debug("parsing overrides params; vars=%s, params=%s, overrides=%s", cfg_vars, cfg_params, overrides)
    configs = ConfigFiles([
        ConfigDocument(parse_override_yaml(overrides)),
        ConfigDocument({"vars": parse_override_list(cfg_vars)}),
        ConfigDocument({"params": parse_override_list(cfg_params)}),
    ])

    log.debug("created overrides: %s", configs)

    return configs

def parse_override_list(overrides: List):
    """
    Parse user-provided cli argument into dict; used to allow user to pass
    var and param overrides on the command-line.
    """
    ret_val = {}
    for kv in overrides:
        key, value = kv.split("=", 1)
        try:
            # try parsing value as a YAML object. E.g. user can do --var 'foo=[123]'
            parsed_value = yaml.safe_load(value)
        except yaml.YAMLError:
            # value passed may not have been yaml, e.g. user may have just done
            # --var foo=bar ; shortcut for having to do --var 'foo="bar"'
            parsed_value = value
        ret_val[key] = parsed_value

    return ret_val

def parse_override_yaml(overrides: Union[str, None]) -> Dict:
    """
    Parse yaml object passed on the command-line

    Can be a yaml-esque string, or @filename, to load
    """
    if overrides:
        if overrides.startswith('@'):
            filename = overrides[1:]
            with open(filename, 'r', encoding="utf-8") as fh:
                return yaml.safe_load(fh)

        return yaml.safe_load(overrides)



from __future__ import annotations

import hashlib
import re

from typing import List, Union
from cfn_tools import load_yaml
from jinja2 import Environment, StrictUndefined, nodes
from jinja2.ext import Extension
from jinja2.exceptions import TemplateRuntimeError

from . import log
from .provider import GenericProvider
from .config import Config
from .template_helpers import TemplateHelpers
from .cfn_bucket import CfnBucket, Uploadable


class FailedTemplate(dict):
    ERROR_CONTEXT_LINES = 3

    def __init__(self, name: str, source: str, location: str = "unknown", error: Exception = None):
        self.name = name
        self.source = source
        self.location = location
        self.error = error

    def __str__(self) -> str:
        """
        Return traceback showing location in template that triggered the exception
        """
        traceback = self.error.__traceback__
        last_frame = traceback
        while traceback:
            filename = traceback.tb_frame.f_code.co_filename
            if filename == "<template>" or filename == "<unknown>":
                break
            last_frame = traceback
            traceback = traceback.tb_next

        # If don't find filename == <template> then exception likely triggered by something
        # outside of template.
        if not traceback:
            line_no = last_frame.tb_lineno
            filename = last_frame.tb_frame.f_code.co_filename
            return f"Error occurred outsite of template\n{str(self.error)}\n{filename}:{line_no}"

        line_no = traceback.tb_lineno
        return f"{str(self.error)}\n{self.location} at line {line_no}:\n\n{self.source_context(line_no)}\n\n"

    def source_context(self, line_no: int) -> str:
        lines = self.source.split("\n")

        from_line = max(1, line_no - self.ERROR_CONTEXT_LINES)
        to_line = min(len(lines), line_no + self.ERROR_CONTEXT_LINES + 1)

        code = []
        for i in range(from_line, to_line):
            line = "%4d : %s" % (i, lines[i - 1])
            if i == line_no:
                line = f"[bold red]{line}[/bold red]"
            code.append(line)
        return "\n".join(code)


class RenderedTemplate(dict, Uploadable):
    def __init__(self, name: str, content: str):
        self.content = content
        self.name = name
        self.error = None  # enables if Template().render().error:...

        parsed = load_yaml(content)
        if not hasattr(parsed, "keys"):
            raise Exception(f"Template result is a {type(parsed)}, expected dict")
        self.update(parsed)

    def __str__(self) -> str:
        return self.content

    def md5(self) -> str:
        return hashlib.md5(self.body()).hexdigest()

    def body(self) -> bytes:
        return self.content.encode("utf-8")

    def key(self) -> str:
        return "/".join(["templates", self.name, self.md5() + ".zip"])

    def iam_capabilities(self) -> List[str]:
        """
        Stacks are decorated with iam capability requirements. Can either do:

            # deprecated
            Metadata:
                iam_capabilities: [....]

        or:

            Metadata:
                stack:
                    iam_capabilities: [....]

        """
        metadata = self.get("Metadata", {})
        if "iam_capabilities" in metadata:
            log.warning("Metadata.iam_capabilities is deprecated. Use Metadata.stack.iam_capabilities instead")
            return metadata.get("iam_capabilities", [])
        return metadata.get("stack", {}).get("iam_capabilities", [])


class RaiseExtension(Extension):
    """
    Extension for raising exceptions in Jinja2 template from
    https://stackoverflow.com/questions/21778252/how-to-raise-an-exception-in-a-jinja2-macro
    """
    tags = set(['raise'])

    # See also: jinja2.parser.parse_include()
    def parse(self, parser):
        # the first token is the token that started the tag. In our case we
        # only listen to "raise" so this will be a name token with
        # "raise" as value. We get the line number so that we can give
        # that line number to the nodes we insert.
        lineno = next(parser.stream).lineno

        # Extract the message from the template
        message_node = parser.parse_expression()

        return nodes.CallBlock(
            self.call_method('_raise', [message_node], lineno=lineno),
            [], [], [], lineno=lineno
        )

    def _raise(self, msg, caller):
        raise TemplateRuntimeError(msg)

class Template:
    class TemplateRenderingException(Exception):
        def __init__(self, template):
            super().__init__("Template could not be rendered")
            self.template = template

    def __init__(self, name: str, provider: GenericProvider, helpers: TemplateHelpers = None):
        if type(name) != str:
            raise Exception(f"Invalid template name {name}")

        self.name = name
        self.provider = provider
        self.helpers = helpers

    def render(self, vars: dict, fail_on_error: bool = False) -> Union[RenderedTemplate, FailedTemplate]:
        raw_template = str(self.provider.template(), "utf-8")

        content = None
        env = Environment(line_statement_prefix="##", undefined=StrictUndefined, extensions=['jinja2_strcase.StrcaseExtension', RaiseExtension])

        if self.helpers:
            self.helpers.inject_helpers(env)

        content = None
        # This will fail if rendered template can't be processed via Jinja2 (e.g. undefined variable access etc)
        try:
            content = env.from_string(source=raw_template).render(vars)

            # "@@{{ some expression }}@@" is the same as {{ some expression }}
            #
            # Helps workaround YAML formatting when editing unquoted {{ jinja expression }}
            #
            content = re.sub(r'"@@(.*)@@"', r"\g<1>", content)

            return RenderedTemplate(name=self.name, content=content)
        except (BaseException, ValueError) as ex:
            template = FailedTemplate(name=self.name, source=(content or raw_template), location=str(self.provider), error=ex)
            if fail_on_error:
                raise self.TemplateRenderingException(template)
            return template


class TemplateWithConfig(Template):
    def __init__(self, provider: GenericProvider, config: Config):
        self.vars = config.vars

        helpers = TemplateHelpers(provider=provider, bucket=CfnBucket(config=config.aws), custom_helpers=config.helpers, config=config)

        # Ugly hack alert. Injecting template commit information to vars.deploy() object here
        try:
            head = provider.head()
            if head:
                if "deploy" in config.vars:
                    deploy = config.vars["deploy"]
                    deploy.template_sha = str(head.hexsha)
                    deploy.template_ref = str(provider.git_ref)
                else:
                    log.info("Skipping setting template git info, no deploy object in config.vars")
            else:
                log.info("No head() for template repo")
        except Exception as ex:
            log.warning("Unable to add template git info", exc_info=ex)

        super().__init__(name=config.template_source.name, provider=provider, helpers=helpers)

    def render(self, fail_on_error: bool = False) -> Union[RenderedTemplate, FailedTemplate]:
        return super().render(self.vars, fail_on_error=fail_on_error)



from .aws_config import AwsSettings
from .basic_stack import BasicStack

class StackReference(BasicStack):
    def __init__(self, aws: AwsSettings, name: str):
        # self._status = None
        super().__init__(aws, name)

    def status(self):
        if not hasattr(self, "_status"):
            self._status = super().status()
        return self._status

    def describe_stack(self):
        """
        For references, cache the describe_stack - we're not expecting
        it to change.
        """
        if not hasattr(self, "_describe_stack_result"):
            self._describe_stack_result = super().describe_stack()
        return self._describe_stack_result

    def description(self):
        """
        return a description of the stack (name and region) to help user find
        it in AWS console
        """
        try:
            status = self.status()
        except Exception:
            status = "?"

        return f"\[cfn_name={self.name},region={self.aws.region},status={status}\]"



from __future__ import annotations
from pathlib import Path
from typing import Union

import yaml


from . import log


class ConfigFiles(list):
    def fetch_dict(self, key, environment, defaults: dict = {}):
        ret_val = dict(defaults)

        for config_file in self:
            # Top-level key in file is lowest priority
            try:
                ret_val.update(config_file.get(key, {}))
            except TypeError as ex:
                self.report_error(f"Unable to retrieve top-level key '{key}'", config_file, key, ex)

            # Environment-specific key is higher priority
            try:
                ret_val.update(config_file.environment(environment).get(key, {}))
            except TypeError as ex:
                self.report_error(f"Unable to retrieve {key} from environments.{environment}", config_file, key, ex)

        return ret_val

    def fetch_set(self, key, environment, defaults: list = []):
        ret_val = set(defaults)

        for config_file in self:
            ret_val.update(config_file.get(key, []))
            ret_val.update(config_file.environment(environment).get(key, []))

        return ret_val

    def validate(self, config):
        valid_environments = config.core.environments
        if valid_environments:
            for include in self:
                include.validate(valid_environments)

    def report_error(self, msg: str, config_file: ConfigFile, key: str, err: Exception):
        print(msg + f" while processing {config_file.filename}: {err}")
        exit(-1)

class ConfigObject(dict):
    """Base object representing a source of configuration - multiple ConfigObjects are merged to generate final configuration by ConfigFiles"""
    EXPECTED_KEYS = set(["aws", "core", "environments", "helpers", "includes", "params", "refs", "tags", "template", "vars"])

    # almost all keys can appear under 'environments:'... except environments
    EXPECTED_ENV_KEYS = EXPECTED_KEYS - set(['environments'])

    def __init__(self):
        super().__init__()
        self.filename = "anonymous"
        self["vars"] = {}
        self["params"] = {}
        self["includes"] = []
        self["helpers"] = []
        self["environments"] = {}
        self["refs"] = {}
        self["template"] = {}
        self["tags"] = {}


    def _ensure_valid_keys(self):
        """
        Ensure config file only contains expected keys
        """
        unknown_keys = set(self.keys()) - self.EXPECTED_KEYS

        if unknown_keys:
            raise Exception(f"Config file {self.filename} has unexpected keys: {unknown_keys}")

        for env_name, env_settings in self["environments"].items():
            if env_settings:  # handle case with blank environment
                unknown_env_keys = set(env_settings.keys()) - self.EXPECTED_ENV_KEYS
                if unknown_env_keys:
                    raise Exception(f"Config file {self.filename} environments.{env_name} has unexpected keys: {unknown_env_keys}")

    def environments(self) -> list:
        """
        Return list of environments defined in this config file. Normally the top-level config is
        special in that it defines which environments a stack can be deployed into - even if the
        included configs defined additional environments.
        """
        return list(self["environments"].keys())

    def environment(self, environment) -> dict:
        """
        Returns environment section from a config file. Returns empty dict if not defined, or None.
        """
        return self["environments"].get(environment, None) or {}

    def validate(self, valid_environments):
        if "environments" in self:
            defined_envs = set(self["environments"].keys())
            invalid_envs = defined_envs - set(valid_environments)
            if invalid_envs:
                invalid_envs = ", ".join(invalid_envs)
                valid_envs = ", ".join(valid_environments)
                raise Exception(f"{self.filename} defines environments {invalid_envs}, which are not listed in core.environments ({valid_envs})")

    def _post_init(self, cfg: Union[dict, None]):
        # hack 'template: [ name: 'template_name' } shortcut
        if cfg and "template" in cfg and isinstance(cfg["template"], str):
            cfg["template"] = {"name": cfg["template"]}

        if cfg:
            log.debug("ConfigObject: initializing from %s", cfg)
            self.update(cfg)

        # check that an environment matching one of the reserved keys isn't defined - highly
        # likely that user has fat-fingered their config
        reserved_env_errors = self.EXPECTED_KEYS & set(self["environments"].keys())
        if reserved_env_errors:
            raise Exception(f"{self.filename} has defined environments {reserved_env_errors}; these are reserved")

        self._ensure_valid_keys()



class ConfigDocument(ConfigObject):
    """A static document representing configuration - used to allow user to provide overrides"""
    def __init__(self, config: Union[dict, None]):
        super().__init__()
        self._post_init(config)

class ConfigFile(ConfigObject):
    """A ConfigObject loaded from file (in the config dir)"""
    def __init__(self, filename: str, config_dir: str):
        super().__init__()
        self.config_dir = config_dir
        self.filename = str(self._find_config_file(Path(filename)))

        filepath = Path(config_dir, self.filename)
        try:
            cfg = yaml.safe_load(open(filepath, "r", encoding="utf-8")) or dict()
        except yaml.parser.ParserError as ex:
            log.fatal("Unable to load configuration file %s", filepath, exc_info=ex)
            raise

        self._post_init(cfg)

    def includes(self):
        """
        Returns list of included files (relative to config dir). Files will be given .yml extension
        if they don't have an extension already
        """
        includes = self["includes"]
        if not isinstance(includes, list):
            raise Exception(f"{self.filename} invalid `includes` directive. Expect a list, got a {type(includes)}")

        # Build a list of ['include/file-1.yml', 'include/file-2.yml]...
        include_paths = []
        for included in includes:
            p = self._find_config_file(Path("includes", included))
            include_paths.append(str(p))
        return include_paths


    def load_includes(self, overrides: Union[ConfigFiles, None] = None) -> ConfigFiles:
        """
        Returns list of config objects with highest precedence last (lowest first)
        """
        config_objects = ConfigFiles([include for include in self._load_includes(seen = set()) if include])
        if overrides:
            config_objects.extend(overrides)
        return config_objects

    def _load_includes(self, seen) -> list:
        """
        Recursive loading of includes. Returns a list with lowest-precedence first.

        E.g. if A includes B, and B includes C then a._load_includes() returns [C, B, A]
        """
        includes = []
        for include in self.includes():
            if include not in seen:
                seen.add(include)
                included = ConfigFile(include, self.config_dir)._load_includes(seen)
                includes += included

        includes += [self]

        return includes

    def _find_config_file(self, p: Path):
        if p.suffix:
            return p

        for suffix in [".yaml", ".yml"]:
            p = p.with_suffix(suffix)
            if Path(self.config_dir, p).exists():
                return p

        raise FileNotFoundError(f"{p.with_suffix('')} does not exist (tried .yaml,.yml)")



"""
Provides `Config` class which handles the loading of configuration and
exposing derived state.
"""
from __future__ import annotations
import json

import pathlib
import re
import os

from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from sys import exc_info
from typing import Union

import git

from jinja2 import Environment, StrictUndefined
from rich.table import Table
from yaml import safe_load

from . import ConfigException, log, console, VERSION
from .config_file import ConfigFiles, ConfigFile
from .template_source import TemplateSource
from .aws_config import AwsSettings
from .stack_refs import StackRefs
from .interpolated_dict import InterpolatedDict, InterpolationError

class Config:
    """
    Represents a final, merged, configuration for a stack deployment.
    """

    @dataclass
    class CoreSettings:
        """
        Settings that change the behavior of the cfn command - not directly related
        to templates or deployments.
        """

        # Attributes
        stack_name: str
        encode_params: bool
        environments: list = None

        # DEFAULTS are pre-interpolation values so can't set them via attributes
        DEFAULTS = {
            "encode_params": False,
            "stack_name": "{{ environment }}-{{ name.replace('/', '-') }}"
        }

        # stack name
        valid_stack_name = re.compile("^[a-zA-Z0-9-]+$").match

        def __post_init__(self):
            if type(self.stack_name) != str or not self.valid_stack_name(self.stack_name):
                raise ValueError(f"Stack name {self.stack_name} is invalid. Can contain only alphanumeric characters and hyphens")

    class Vars(dict):
        MAX_INTERPOLATION_DEPTH = 10

        def __init__(self, vars: dict):
            """
            This method initializes an object with a dictionary of variables 'vars'. It expands the interpolated variables in the
            dictionary using Jinja2, and checks for any failed keys. If there are any failed keys, it creates a table of errors with
            the key name, corresponding value, and error message, and logs it to the console. It then raises an exception with a
            message indicating that an error occurred processing the 'vars' dictionary.

            This is a bit different than how to expand other Jinja-value dicts because we'll need to iterate through a few
            times until all vars have been expanded.
            """
            failed_keys = self.expand(vars)

            if failed_keys:
                errors = Table("Key", "Value", "Error")
                for k, v in failed_keys.items():
                    errors.add_row(k, str(v.value), str(v.error))
                console.log(errors)
                raise Exception(f"An error occurred processing vars: {failed_keys}")

        def expand(self, vars: dict):
            """
            Updates `self` with final interpolated values from initial dict `vars`.

            Returns dict InterpolationError for all keys that can't be expanded
            """
            env = Environment(undefined=StrictUndefined, extensions=["jinja2_strcase.StrcaseExtension"])

            interpolation_depth = 0

            errors = {}
            while interpolation_depth <= self.MAX_INTERPOLATION_DEPTH:
                interpolation_depth += 1
                unexpanded_keys = sorted(set(vars.keys()) - set(self.keys()))
                if not unexpanded_keys:
                    return None

                for key in unexpanded_keys:
                    value = vars[key]
                    try:
                        del errors[key]
                        if type(value) in [bool, dict, list, str]:
                            tpl = env.from_string(str(value))  # convert value to jinja2 template
                            result = str(tpl.render(self))
                            self[key] = safe_load(result)
                            # print(f"expanding {key}, {value}({type(value)}) -> {self[key]}({type(self[key])})")
                        else:
                            # print(f"skipping {key}, type={type(value)}")
                            self[key] = value  # Don't try and process this value as Jinja template
                        del vars[key]
                    except Exception:
                        errors[key] = InterpolationError(key, value, exc_info()[1])

            return errors

    class Tags(InterpolatedDict):
        def to_list(self, extra_attributes={}):
            ret_val = []
            for k, v in self.items():
                ret_val.append({"Key": str(k), "Value": str(v), **extra_attributes})
            return ret_val

    @dataclass
    class DeployMetadata:
        def __init__(self, config_path: str, template_source: TemplateSource):
            # Timestamp in UTC
            self.timestamp = datetime.utcnow().strftime("%Y-%m-%d-%H:%M:%S%Z")

            # STK version
            self.deployed_with = f"stk-{VERSION}"

            root = template_source.root or ""
            name = template_source.name or ""

            if template_source.repo:
                self.template = "/".join([template_source.repo, root, name])
            else:
                self.template = "/".join([root, name])

            # Config git HEAD state - this is optional. The config project may
            # not be stored in git. Also, if config is in a subdirectory, we'd
            # too dumb to try and find git repo in a parent directory.
            try:
                config_head = git.Repo(config_path).head
                self.config_sha = str(config_head.commit.hexsha)
                self.config_ref = str(config_head.reference)
            except Exception as ex:  # pylint: disable=broad-except
                log.debug("Unable to retrieve git info for config project", exc_info=ex)
                self.config_sha = "?"
                self.config_ref = "?"

            self.template_sha = "?"
            self.template_ref = "?"

    def __init__(
        self,
        name: str,
        environment: str,
        config_path: str,
        overrides: ConfigFiles = ConfigFiles([]),
        template_path: Union[str, None] = None,
    ):
        # While we should just receive `name`, we may be be passed
        name = str(Path(name).with_suffix(""))
        self.name = name.replace("/", "-")
        self.environment = environment
        self.config_path = config_path

        try:
            cfg = ConfigFile(filename=name, config_dir=self.config_path)
            log.debug("loaded initial config file %s from %s: %s", name, self.config_path, cfg, extra={"cfg": cfg})
        except FileNotFoundError as err:
            raise Exception(f"Configuration file {name} not found in {config_path}: {err}") from err

        # Validate specified environment is defined in the top-level config file
        if environment not in cfg.environments():
            raise ConfigException(f"Environment {environment} is not a valid environment for {cfg.filename}. Only {cfg.environments()} permitted.")

        # Load top-level config file and all included configs
        includes = cfg.load_includes(overrides)

        # Most other config supports pulling stuff from AWS, so initialize this first
        aws_settings = {}
        try:
            aws_settings = InterpolatedDict(includes.fetch_dict("aws", environment), {"environ": os.environ, "environment": environment})
            self.aws = AwsSettings(**aws_settings)
            self.aws.get_account_id()  # force retrieval of account_id
        except TypeError as ex:
            raise Exception(f"Unable to parse aws settings: have {aws_settings}: {ex}") from ex

        default_vars = {
            "__config_dir": pathlib.Path(config_path),
            "account_id": self.aws.account_id,
            "aws_region": self.aws.region,
            "cfn_bucket": self.aws.cfn_bucket,
            "environ": os.environ,
            "environment": environment,
            "name": self.name,
        }

        # Core settings impact the behavior of 'stk' - e.g. stack name, valid environments
        # etc.
        self.core = self.CoreSettings(
            **InterpolatedDict(
                includes.fetch_dict("core", environment, self.CoreSettings.DEFAULTS),
                default_vars,
            )
        )

        # Ugly hack. Need to come up with something better after I've had a coffee
        default_vars["stack_name"] = self.core.stack_name

        # Stack 'refs' object references external stacks. They are intended to be resolved by 'vars'/'params' so need to be
        # loaded first
        try:
            refs = InterpolatedDict(includes.fetch_dict("refs", environment), {"environment": environment})
            self.refs = StackRefs(refs, self)
        except Exception as ex:
            raise Exception("Unable to parse stack refs (refs:). have {refs}: {ex}") from ex
        default_vars["refs"] = self.refs

        self.helpers = list(includes.fetch_set("helpers", environment))

        pre_vars = includes.fetch_dict("vars", environment, default_vars)
        self.vars = self.Vars(pre_vars)

        params = includes.fetch_dict("params", environment)
        self.params = InterpolatedDict(params, self.vars)
        if self.core.encode_params:
            self.encode_param_values(self.params)
        log.debug("setting parameters: %s", self.params)
        self.vars["params"] = self.params

        template_source = InterpolatedDict(
            includes.fetch_dict(
                "template",
                environment,
                {"name": name.replace("/", "-"), "root": None},
            ),
            self.vars,
        )

        # I'm not happy about this
        if template_source["root"] == None:
            if "repo" in template_source and template_source["repo"]:
                # git repo, we default "root" to / - i.e. relevant to git root
                template_source["root"] = "/"
            else:
                # filesystem repo, we default "root" to provided template_path (--template-path args)
                # mostly useful for tests
                template_source["root"] = template_path

        self.template_source = TemplateSource(**template_source)

        # Deploy metadata is used to track deploys back to version controlled config/templates.
        self.vars["deploy"] = self.DeployMetadata(config_path=config_path, template_source=self.template_source)

        self.tags = self.Tags(includes.fetch_dict("tags", environment), self.vars)

        # perform final linting/validation
        includes.validate(self)

    def var(self, name):
        return self.vars.get(name)

    def param(self, name):
        return self.params.get(name)

    def encode_param_values(self, params):
        """
        JSON-encode any CFN *parameters* that aren't strings. Enabled by
        setting `core.encode_params`. Useful if you want to pass structured
        data as JSON to the template (e.g. setting parameters, or secrets).
        """
        for k, value in params.items():
            if not isinstance(value, str):
                params[k] = json.dumps(value)



from __future__ import annotations

import base64
import contextlib
import hashlib
import json
import os
import re
import shutil
import tempfile
import time

from dataclasses import dataclass
from importlib import util as importutil
from os import path
from pathlib import Path
from typing import IO
from zipfile import ZipFile, ZipInfo, ZIP_DEFLATED

import jinja2

from jinja2 import Environment

from . import log
from .human_bytes import HumanBytes
from .cfn_bucket import CfnBucket, Uploadable
from .ignore_file import parse_ignore_list
from .multipart_encoder import multipart_encode
from .config import Config
from .provider import GenericProvider


@dataclass
class ZipContent(Uploadable):
    name: str
    content: IO
    md5sum: str

    def body(self) -> IO:
        return self.content

    def key(self) -> str:
        return "/".join([self.name, self.md5sum + ".zip"])


@contextlib.contextmanager
def in_tmp_directory():
    temp_dir = tempfile.mkdtemp()
    try:
        saved_dir = os.getcwd()
        os.chdir(temp_dir)
        yield
    finally:
        os.chdir(saved_dir)
        shutil.rmtree(temp_dir)


class TemplateHelpers:
    def __init__(self, provider: GenericProvider, bucket: CfnBucket, custom_helpers: list, config: Config):
        self.provider = provider
        self.bucket = bucket
        self.config = config
        self.custom_helpers = {}

        # These are "short-cuts" for use by custom helpers
        self.aws = config.aws

        with in_tmp_directory():
            os.mkdir("helpers")
            for name in custom_helpers:
                self.custom_helpers[name] = self._load_custom_helper(name)

    def inject_helpers(self, env: Environment):
        """
        Inject helpers into a Jinja2 environment. We have a set of core (standard) helpers that
        should be useful for most projects.

        Template projects can define custom helpers for domain-specific logic.
        """
        log.debug("injecting helpers into jinja environment")
        funcs = env.globals

        # Core helpers
        funcs["resourcify"] = self.resourcify
        funcs["lambda_uri"] = self.lambda_uri
        funcs["lambda_code"] = self.lambda_code
        funcs["user_data"] = self.user_data
        funcs["include_file"] = self.include_file
        funcs["upload_zip"] = self.upload_zip
        funcs["tags"] = self.tags
        funcs["resource_cidr"] = self.resource_cidr
        funcs["b64encode"] = self.b64encode

        # Core filters
        filters = env.filters
        filters['b64encode'] = self.b64encode

        # Custom helpers (defined in templates/helpers and specified in config via 'helpers' stanza)
        for name, func in self.custom_helpers.items():
            # Do this via func to avoid late binding problems
            # https://stackoverflow.com/questions/3431676/creating-functions-in-a-loop
            funcs[name] = self._make_helper_wrapper(func)

    def _make_helper_wrapper(self, func):
        return lambda *args, **kwargs: func(self, *args, **kwargs)

    def b64encode(self, s: str) -> str:
        """Base64 encode a string, used both as a filter and a function"""
        return base64.b64encode(s.encode('utf-8')).decode('utf-8')

    def resourcify(self, name) -> str:
        """
        Given a string with non-alphanumeric characters, maps to a string that can be used as an AWS Resource name.
        """
        return re.sub(r"(\A|\W)+(\w)", lambda m: m.group(2).upper(), re.sub(r"(_|-)+", " ", str(name))).replace(" ", "")

    IGNORE_FILE = ".package-ignore"

    def lambda_uri(self, name: str, prefix: str="") -> str:
        """Upload lambda to S3 and return an https URI to the code"""
        lambda_path = path.join("functions", name)

        zipped = self.zip_tree(dir=lambda_path, ignore=self.ignore_list(lambda_path), prefix=prefix)
        return self.bucket.upload(zipped).as_s3()

    def lambda_code(self, name: str, prefix: str="") -> dict:
        """Upload lambda to S3 and return a {S3Bucket, S3Key} pair to the result"""
        lambda_path = path.join("functions", name)

        zipped = self.zip_tree(dir=lambda_path, ignore=self.ignore_list(lambda_path), prefix=prefix)
        uploaded = self.bucket.upload(zipped)

        return {"S3Bucket": uploaded.bucket.bucket_name, "S3Key": uploaded.key}

    def ignore_list(self, p: str):
        provider = self.provider

        # Ignore files in $TEMPLATE_ROOT/$type/$name or $TEMPLATE_ROOT/$type
        dir = Path(p)
        ignore_files = [str(dir.parent / self.IGNORE_FILE), str(dir / self.IGNORE_FILE)]

        # Make a merged ignore list (and also add the .package-ignore to the list)
        ignore_content = "\n".join([self.IGNORE_FILE] + [str(provider.content(p), "utf-8") for p in ignore_files if provider.is_file(p)])

        # Parse final list
        return parse_ignore_list(ignore_content)

    def user_data(self, name: str, **extra_vars) -> str:
        """
        Given a named user_data/<name> directory, generates UserData content
        """
        dir = path.join("user_data", name.lower())
        log.debug("adding user_data from %s", dir)

        if not self.provider.is_tree(dir):
            raise (Exception(f"{dir} is not a directory"))

        # Jinja2 evaluation requires any referenced variables be defined
        # to avoid hard-to-detect failures.
        log.debug("setting up jinja environment")
        env = jinja2.Environment(line_statement_prefix="##", undefined=jinja2.StrictUndefined)
        self.inject_helpers(env)

        # context for template evaluation is 'config.vars' plus any additional
        # parameters passed. E.g.
        #
        #     user_data(name='foo', param1='bar', param2='buzz')
        #
        template_context = {**self.config.vars, **extra_vars}

        # Build dict of { name => content } that we can encode
        parts = {}
        for part_name, content_type, content in self.provider.find(dir):
            log.debug("processing part %s type=%s", part_name, content_type)
            if content_type != "file":
                raise Exception("user_data(): %s is not a regular file" % part_name)

            # Userdata files are actually Jinja2 templates in disguise
            template = env.from_string(source=str(content, "utf-8"))
            log.debug("rendering template for part %s", part_name)
            parts[part_name] = template.render(template_context)

        log.debug("performing multipart encode")
        encoded = multipart_encode(sorted(parts.items()))

        # Map encoded multi-line string to a JSON array.
        #
        # To allow resource references within user data, json fragments
        # can be embedded within << >> delimeters.
        #
        # for example: "hello <<{"Ref": "bar"}>> there <<{}>>" will be mapped
        # to a json array ["hello ", {"Ref": "bar"}, " there ", {}]
        #
        log.debug("processing break-outs")
        lines = []
        for line in encoded.splitlines(keepends=True):
            parts = re.split("<<(.+?)>>(?!>)", line)
            for i, match in enumerate(parts):
                if i % 2:
                    log.debug("processing json block %s", match)
                    lines.append(json.loads(match))
                else:
                    lines.append(match)

        # Rendering user data as correctly indented content is hard, so
        # don't even bother - just dump out single-line JSON !
        log.debug("user_data complete")
        return json.dumps({"Fn::Base64": {"Fn::Join": ["", lines]}})

    def tags(self, extra_attributes={}, **tags):
        final_tags = Config.Tags({**self.config.tags, **tags}, {})
        return final_tags.to_list(extra_attributes=extra_attributes)

    def resource_cidr(self, resource_id):
        """
        Find CIDR for subnet or VPC (AWS resources with a CIDR associated), otherwise
        just returns resource_id.
        """
        ec2 = self.aws.resource("ec2")
        if resource_id.startswith("subnet-"):
            return ec2.Subnet(resource_id).cidr_block
        elif resource_id.startswith("vpc-"):
            return ec2.Vpc(resource_id).cidr_block
        else:
            # Maybe passed in a CIDR already
            return resource_id

    def include_file(self, include_file_name, padding=8, prefix="\n", **extra_vars) -> str:
        env = Environment(line_statement_prefix="##", undefined=jinja2.StrictUndefined)

        self.inject_helpers(env)

        content = self.provider.content(path.join("files", include_file_name))
        template = env.from_string(source=str(content, "utf-8"))

        # context for evalation is 'config.vars' plus any additional parameters passed via
        # :extra_vars:
        result = template.render({**self.config.vars, **extra_vars})

        indended = "\n".join(map(lambda line: " " * padding + line, result.splitlines())) + "\n"

        return prefix + indended

    def upload_zip(self, dir: str, prefix: str = "", returns: str = "key") -> str:
        zipped = self.zip_tree(dir=dir, prefix=prefix)

        res = self.bucket.upload(zipped)
        if returns == "key":
            return res.key
        elif returns == "s3-uri":
            return res.as_s3()
        elif returns == "http-uri":
            return res.as_http()
        else:
            raise Exception(f"Unknown 'returns' value {returns} - expect one of (key, s3-uri, http-uri)")

    def zip_tree(self, dir: str, ignore=None, prefix="") -> ZipContent:
        """
        Compress directory tree (root), setting prefix for files inside zip
        """
        checksums = {}

        tmp_file = tempfile.TemporaryFile()
        with ZipFile(tmp_file, mode="w", compression=ZIP_DEFLATED) as zip:
            log.info(f"Adding files from {dir}")
            count, size = 0, 0
            for file_path, type, file_content in self.provider.find(dir, ignore):
                file_size = len(file_content)
                log.debug(" adding file %s (size=%s, type=%s)", HumanBytes.format(file_size), file_path, type)
                file_path = path.join(prefix, file_path)

                # Add file to zip
                info = ZipInfo(filename=file_path, date_time=time.localtime(time.time())[:6])
                info.file_size = file_size

                # Set file perm ugo=rx, preserve symlinks - 0xa000 (0x120000) bit
                info.external_attr = (0o120755 if type == "symlink" else 0o555) << 16

                zip.writestr(info, data=file_content, compress_type=ZIP_DEFLATED)

                # Record md5 checksum
                checksums[path.join(dir, file_path)] = hashlib.md5(str(file_content).encode("utf-8")).hexdigest()

                count += 1
                size += info.file_size

            log.info("Added %d files, total %s", count, HumanBytes.format(size))

        # final (composite) checksum is based on filenames and content md5s. They are sorted so checksum doesn't
        # vary if files are discovered in different orders.
        sorted_checksums = sorted(checksums.items())
        md5sum = hashlib.md5(str(sorted_checksums).encode("utf-8")).hexdigest()

        tmp_file.seek(0)  # required?

        return ZipContent(dir, tmp_file, md5sum)

    def _load_custom_helper(self, name: str):
        """
        TBH I don't really understand this, stolen from stack overflow 😱
        """
        mod_name = f"stk.template_helpers.config.{name}"

        mod_file = str(Path("helpers", name).with_suffix(".py"))

        log.debug("retrieving helper %s from provider", mod_file)
        content = self.provider.content(mod_file)

        log.debug("writing helper code to %s", mod_file)
        open(mod_file, "wb").write(content)

        spec = importutil.spec_from_file_location(mod_name, mod_file)

        if spec is None:
            raise ImportError(f"Could not load spec for module '{mod_name}' at: {mod_file}")

        module = importutil.module_from_spec(spec)
        try:
            spec.loader.exec_module(module)
        except FileNotFoundError as e:
            raise ImportError(f"{e.strerror}: {mod_file}") from e

        helper_func = getattr(module, "helper")

        return helper_func



# pylint: disable=missing-docstring

import json

from pytest import fixture, raises, mark
from botocore.stub import Stubber

from . import ConfigFixtures, StackFixtures
from ..config import Config
from ..stack import Stack
from ..template import RenderedTemplate

@mark.skip("moto outputs seem to be broken at the moment")
class TestStackRefs(ConfigFixtures, StackFixtures):
    @fixture
    def config(self, sts):
        return Config("main", environment="dev", config_path=self.fixture_path("refs", "config"))

    @fixture
    def required_stack(self, cloudformation, cfn_bucket, aws):
        """
        Create dev-some-other-stack from fixtures/refs/templates/outputs.yaml
        """

        # Render template
        template = RenderedTemplate(name="outputs.yaml", content=open(self.fixture_path("refs", "templates", "outputs.yaml"), "r").read())

        # Create stack from template
        stack = Stack(aws=aws, name="dev-some-other-stack")
        result = stack.create_change_set(template=template).execute()
        return stack

    def test_config_parses(self, required_stack, config):
        refs = config.refs
        assert not refs.stack("optional_stack").exists()
        assert refs.stack("required_stack").exists()
        assert refs.output("optional_stack", "foo") == None

        assert not refs.exists("optional_stack")
        assert refs.exists("required_stack")

    def test_stack_not_defined(self, config):
        v = Config.StackRefs(stack_refs={}, config=config)
        with raises(Exception, match="Attempt to access stack foo, but it's not defined in config.refs"):
            v.stack("foo")

    def test_stack_not_exists_with_defaults(self, config, cloudformation):
        v = Config.StackRefs(stack_refs={"foo": None}, config=config)
        with raises(Exception, match="Stack config.refs\[foo\] \(dev-foo\) does not exist, but is required"):
            v.output("foo", "bar")

    def test_stack_not_exists_with_name(self, config, cloudformation):
        v = Config.StackRefs(stack_refs={"foo": {"stack_name": "something-{{ environment }}-else"}}, config=config)
        with raises(Exception, match="Stack config.refs\[foo\] \(something-dev-else\) does not exist, but is required"):
            v.output("foo", "bar")

    def test_multiple_required_stacks_not_exists(self, config, cloudformation):
        refs = Config.StackRefs(stack_refs={"foo": {"stack_name": "something-{{ environment }}-else"}, "bar": None, "buz": {}}, config=config)

        assert not refs.stack("foo").exists()
        assert not refs.stack("bar").exists()
        assert not refs.stack("buz").exists()

        with raises(Exception, match="Stack config.refs\[foo\] \(something-dev-else\) does not exist, but is required"):
            refs.output("foo", "some_output")

        with raises(Exception, match="Stack config.refs\[bar\] \(dev-bar\) does not exist, but is required"):
            refs.output("bar", "some_output")

    def test_invalid_stack_ref(self, config, cloudformation):
        refs = Config.StackRefs(
            stack_refs={
                "foo": {"stack_name": "something-{{ environment }}-else"},
                "optional": True,
            },
            config=config,
        )

        with raises(SystemExit, match="-1"):
            refs.output("foo", "some_output")

    def test_optional_stacks_not_exists(self, config, cloudformation):
        refs = Config.StackRefs(
            stack_refs={
                "foo": {"stack_name": "something-{{ environment }}-else", "optional": True},
                "bar": {"stack_name": "something-{{ environment }}-else", "optional": "true"},
            },
            config=config,
        )

        assert refs.output("foo", "imaginary_output") == None
        assert refs.output("bar", "imaginary_output") == None

    def test_stack_exist_no_output(self, config, required_stack):
        refs = Config.StackRefs(stack_refs={"some-other-stack": None}, config=config)

        with raises(KeyError, match="'unknown_output not in outputs of dev-some-other-stack"):
            refs.output("some-other-stack", "unknown_output")

    def create_cfn_mock_response(self, refs):
        # refs.stacks()
        for name in refs.stacks().keys():
            # mock the cfn client
            stubber = Stubber(refs[name].cfn)
            expected_params = {"StackName": f"dev-{name}"}
            response = json.loads(self.fixture_content("mock-describe-stacks-response.json"))
            stubber.add_response("describe_stacks", response, expected_params)

            return stubber

    def test_output_not_exists(self, config, cloudformation):
        """
        Accessing a non-existant output on a stack will raise an exception
        """
        refs = Config.StackRefs(stack_refs={"some-other-stack": None}, config=config)
        stubber = self.create_cfn_mock_response(refs)
        with stubber:
            with raises(Exception, match="AnOutput not in outputs of dev-some-other-stack - only have \['MyOutput'\]"):
                refs.output("some-other-stack", "AnOutput")

    def test_output_exist(self, config, cloudformation):
        """
        Referencing an output that *does* exist will actually return the value :-)
        """
        refs = Config.StackRefs(stack_refs={"some-other-stack": None}, config=config)
        stubber = self.create_cfn_mock_response(refs)
        with stubber:
            output = refs.output("some-other-stack", "MyOutput")
            assert output == "An Output Value"
            assert output.description == "Some description"



from __future__ import annotations
from dataclasses import dataclass

from .provider import provider


@dataclass
class TemplateSource:
    """
    Template source defines the origin/source of templates (and associated files).

    We support:
    * Local filesystem
    * Local git repository
    * Remote git repository
    """

    name: str
    """name of root template (without .yaml/.yml extension)"""

    root: str = None
    """path to file - either local, or relative to git root if using repo"""

    version: str = None
    """if set, template is stored in git repository (either local or remote)"""

    repo: str = None
    """required git url if version specified. Ignored if :version: is blank/None"""

    def provider(self):
        return provider(self)



import os
import pytest

# pylint: disable=missing-docstring

from . import Fixtures
from ..aws_config import AwsSettings


class TestMinimalAwsConfig(Fixtures):
    def test_minimal(self):
        aws = AwsSettings(region="us-west-2", cfn_bucket="stk-cfn")
        assert aws.region == "us-west-2"
        assert aws.cfn_bucket == "stk-cfn"



"""
Provides TemplateCommand
"""

from . import console
from .change_set import ChangeSet
from .stack_delegated_command import StackDelegatedCommand
from .template import TemplateWithConfig


class TemplateCommand(StackDelegatedCommand):
    """
    Command that applies a template to a stack
    """

    def __post_init__(self):
        super().__post_init__()
        self.template = TemplateWithConfig(
            provider=self.config.template_source.provider(), config=self.config).render()

        parse_error = self.template.error
        if parse_error:
            console.log(
                f":x: Template is NOT ok - {parse_error}", emoji=True, style="red")
            console.print(str(self.template))
            exit(-1)

    def validate(self):
        """
        Validate template by parsing it (using provided config) and applies
        AWS CFN validation.
        """
        template = self.template
        if not template.error:
            errors = self.stack.validate(template)
            if errors:
                console.print(str(self.template))
                console.log(f"{errors}\n\n", style="red", markup=False)
                console.log(
                    ":x: Template is NOT ok - failed validation", emoji=True, style="red")
                exit(-1)
            else:
                console.log(":+1: Template is ok", emoji=True, style="green")

    def create_change_set(self, change_set_name=None) -> ChangeSet:
        """
        Create a change set
        """
        with console.status("Creating change set"):
            tags = self.config.tags.to_list()
            params = self.config.params
            cs = self.stack.create_change_set(
                template=self.template,
                tags=tags,
                change_set_name=change_set_name,
                params=params
            )
        return cs



from __future__ import annotations
from typing import List

from botocore.exceptions import WaiterError
from rich.table import Table
from rich import box

from . import log
from .stack_waiter import StackWaiter
from .template import RenderedTemplate
from .basic_stack import BasicStack


class ChangeSet:
    ROW_STYLES = {"Add": "green", "Remove": "red", "Modify": "yellow"}

    def __init__(self, stack: BasicStack, name: str):
        self.stack = stack
        self.name = name
        self.cfn = stack.cfn

    def create(self, template: RenderedTemplate, change_set_type: str, tags: List(str), params: dict):
        stack = self.stack
        template_url = stack.bucket.upload(template).as_http()

        parameters = []
        for key, value in params.items():
            parameters.append({"ParameterKey": key, "ParameterValue": value})

        log.debug("creating change set", extra={"parameters": parameters})

        self.cfn.create_change_set(
            StackName=stack.name,
            TemplateURL=template_url,
            Parameters=parameters,
            ChangeSetName=self.name,
            ChangeSetType=change_set_type,
            Capabilities=template.iam_capabilities(),
            Tags=tags,
        )

        try:
            self.wait("change_set_create_complete")
        except WaiterError as ex:
            pass

        return self

    def execute(self):
        self.cfn.execute_change_set(**self._descriptor())
        return self

    def delete(self):
        res = self.cfn.delete_change_set(**self._descriptor())
        return res["ResponseMetadata"]["HTTPStatusCode"] == 200

    def available(self, change_set=None) -> bool:
        status = self._change_set(change_set).get("ExecutionStatus", None)

        return status == "AVAILABLE"

    def is_empty_changeset(self, change_set=None) -> bool:
        """
        Returns true if looks like this change set failed due as a result of
        no changes (not really a failure)

        TODO: Pretty hacky and not language safe
        """
        cs = self._change_set(change_set)

        ex_status = cs.get("ExecutionStatus", "-")
        status = cs.get("Status", "")
        reason = cs.get("StatusReason", "")

        return \
            ex_status == "UNAVAILABLE" and \
            status == "FAILED" and ( \
                "The submitted information didn't contain changes" in reason or
                "No updates are to be performed" in reason
            )


    def _change_set(self, change_set=None):
        if not change_set:
            change_set = self.cfn.describe_change_set(**self._descriptor())
        return change_set

    def wait(self, waiter_name: str):
        StackWaiter(self.stack).wait_for_change_set(waiter_name, self)

    def summary(self) -> Table:
        change_set = self._change_set()

        if self.available(change_set):
            detail = Table("Resource", "Type", "Action", "Details", box=box.SIMPLE)
            for chg in self.changes(change_set):
                row_style = self.ROW_STYLES.get(chg["Action"], "")
                detail.add_row(chg["LogicalResourceId"], chg["ResourceType"], chg["Action"], self.humanize_change_detail(chg.get("Details")), style=row_style)
            return detail

        summary = Table("Property", "Value", title="Change Set", box=box.SIMPLE)
        summary.add_row("Stack Name", change_set.get("StackName", "-"))
        summary.add_row("Stack ID", change_set.get("StackId", "-"))
        summary.add_row("Execution Status", change_set.get("ExecutionStatus", "-"))
        summary.add_row("Status", change_set.get("Status", "-"))
        summary.add_row("Reason", change_set.get("StatusReason", "-"))

        return summary

    def changes(self, cs=None):
        if not cs:
            id = self._descriptor()
            cs = self.cfn.describe_change_set(**self._descriptor())

        if "Changes" in cs:
            for change in cs["Changes"]:
                if "ResourceChange" in change:
                    rs = change["ResourceChange"]
                    yield (rs)

    def resources(self) -> dict:
        """
        Returns list of resources and types in change set. This is used to pre-populate status
        table when executing the change set.

        :returns: dict<LogicalResourceId,ResourceType>
        """
        ret_val = {}
        for rc in self.changes():
            ret_val[rc["LogicalResourceId"]] = rc["ResourceType"]
        return ret_val

    def humanize_change_detail(self, change_details: List[dict]) -> str:
        ret_val = []
        for change in change_details or []:
            try:
                target = change["Target"]
                if "Name" in target:
                    target_name = "[b]%s.%s[/b]" % (target["Attribute"], target["Name"])
                    if change["ChangeSource"] == "DirectModification":
                        ret_val.append(f"{target_name} changed")
                    elif change["ChangeSource"] == "ResourceAttribute":
                        ret_val.append(f"{target_name} changed by {change['CausingEntity']}")
                    else:
                        ret_val.append(str(change))
                else:
                    ret_val.append("[b]%s[/b]" % (target["Attribute"]))
            except Exception as ex:
                print(ex)
                ret_val.append(str(change))
        return "\n".join(ret_val)

    def _descriptor(self) -> dict:
        id = {"StackName": self.stack.name, "ChangeSetName": self.name}
        if not hasattr(self, "id"):
            change_set = self.cfn.describe_change_set(**id)
            self.id = change_set["ChangeSetId"]

        return {"ChangeSetName": self.id}



# Stolen from https://github.com/mherrmann/gitignore_parser

import os
import re

from dataclasses import dataclass
from types import FunctionType


def parse_ignore_list(ignore_content: str) -> FunctionType:
    if type(ignore_content) == bytes:
        ignore_content = str(ignore_content, "utf-8")

    rules = []
    for line in ignore_content.splitlines():
        rule = rule_from_pattern(line)
        if rule:
            rules.append(rule)

    return lambda file_path: is_ignored(file_path, rules)


def is_ignored(file_path: str, rules: FunctionType) -> bool:
    matched = False
    for rule in rules:
        if rule.match(file_path):
            matched = not rule.negation
    return matched


def rule_from_pattern(pattern):
    """
    Take a .gitignore match pattern, such as "*.py[cod]" or "**/*.bak",
    and return an IgnoreRule suitable for matching against files and
    directories. Patterns which do not match files, such as comments
    and blank lines, will return None.
    Because git allows for nested .gitignore files, a base_path value
    is required for correct behavior. The base path should be absolute.
    """

    # Store the exact pattern for our repr and string functions
    orig_pattern = pattern

    # Early returns follow
    # Discard comments and separators
    if pattern.strip() == "" or pattern[0] == "#":
        return

    # Discard anything with more than two consecutive asterisks
    if pattern.find("***") > -1:
        return

    # Strip leading bang before examining double asterisks
    if pattern[0] == "!":
        negation = True
        pattern = pattern[1:]
    else:
        negation = False

    # Discard anything with invalid double-asterisks -- they can appear
    # at the start or the end, or be surrounded by slashes
    for m in re.finditer(r"\*\*", pattern):
        start_index = m.start()
        if start_index != 0 and start_index != len(pattern) - 2 and (pattern[start_index - 1] != "/" or pattern[start_index + 2] != "/"):
            return

    # Special-casing '/', which doesn't match any files or directories
    if pattern.rstrip() == "/":
        return

    directory_only = pattern[-1] == "/"
    # A slash is a sign that we're tied to the base_path of our rule
    # set.
    anchored = "/" in pattern[:-1]
    if pattern[0] == "/":
        pattern = pattern[1:]
    if pattern[0] == "*" and len(pattern) >= 2 and pattern[1] == "*":
        pattern = pattern[2:]
        anchored = False
    if pattern[0] == "/":
        pattern = pattern[1:]
    if pattern[-1] == "/":
        pattern = pattern[:-1]
    # patterns with leading hashes are escaped with a backslash in front, unescape it
    if pattern[0] == "\\" and pattern[1] == "#":
        pattern = pattern[1:]
    # trailing spaces are ignored unless they are escaped with a backslash
    i = len(pattern) - 1
    striptrailingspaces = True
    while i > 1 and pattern[i] == " ":
        if pattern[i - 1] == "\\":
            pattern = pattern[: i - 1] + pattern[i:]
            i = i - 1
            striptrailingspaces = False
        else:
            if striptrailingspaces:
                pattern = pattern[:i]
        i = i - 1

    regex = fnmatch_pathname_to_regex(pattern, directory_only)

    if anchored:
        regex = "".join(["^", regex])

    return IgnoreRule(
        pattern=orig_pattern,
        regex=regex,
        negation=negation,
        directory_only=directory_only,
        anchored=anchored,
    )


whitespace_re = re.compile(r"(\\ )+$")


@dataclass
class IgnoreRule:
    pattern: str
    regex: re.Pattern
    negation: bool
    directory_only: bool
    anchored: bool

    def __str__(self):
        return self.pattern

    def __repr__(self):
        return "".join(["IgnoreRule('", self.pattern, "')"])

    def match(self, rel_path):
        """
        Returns true if rel_patch is matched by this rule
        """
        matched = False
        if rel_path.startswith("./"):
            rel_path = rel_path[2:]

        if re.search(self.regex, rel_path):
            matched = True

        return matched


# Frustratingly, python's fnmatch doesn't provide the FNM_PATHNAME
# option that .gitignore's behavior depends on.
def fnmatch_pathname_to_regex(pattern, directory_only: bool):
    """
    Implements fnmatch style-behavior, as though with FNM_PATHNAME flagged;
    the path separator will not match shell-style '*' and '.' wildcards.
    """
    i, n = 0, len(pattern)

    seps = [re.escape(os.sep)]
    if os.altsep is not None:
        seps.append(re.escape(os.altsep))

    seps_group = "[" + "|".join(seps) + "]"
    nonsep = r"[^{}]".format("|".join(seps))

    res = []
    while i < n:
        c = pattern[i]
        i += 1
        if c == "*":
            try:
                if pattern[i] == "*":
                    i += 1
                    res.append(".*")
                    if pattern[i] == "/":
                        i += 1
                        res.append("".join([seps_group, "?"]))
                else:
                    res.append("".join([nonsep, "*"]))
            except IndexError:
                res.append("".join([nonsep, "*"]))
        elif c == "?":
            res.append(nonsep)
        elif c == "/":
            res.append(seps_group)
        elif c == "[":
            j = i
            if j < n and pattern[j] == "!":
                j += 1
            if j < n and pattern[j] == "]":
                j += 1
            while j < n and pattern[j] != "]":
                j += 1
            if j >= n:
                res.append("\\[")
            else:
                stuff = pattern[i:j].replace("\\", "\\\\")
                i = j + 1
                if stuff[0] == "!":
                    stuff = "".join(["^", stuff[1:]])
                elif stuff[0] == "^":
                    stuff = "".join("\\" + stuff)
                res.append("[{}]".format(stuff))
        else:
            res.append(re.escape(c))

    res.insert(0, "(?ms)")

    if not directory_only:
        res.append("$")

    return "".join(res)



from copy import copy
from dataclasses import dataclass
from typing import Any, Dict, Union

from . import log
from .stack_reference import StackReference
from .interpolated_dict import InterpolatedDict


@dataclass
class StackRefOpts:
    """
    Data class representing a stack reference. User defines config like:

    refs:
        foo_bar:
        some_other_stack:
            stack_name: "override_stack_name"
            region: ap-southeast-3

    """
    stack_name: str
    region: str
    optional: bool

class OptionalStackReference:
    def __init__(self, ref_name: str, stack: StackReference, optional: bool):
        self.stack = stack
        self.ref_name = ref_name
        self.optional = optional

    def __getitem__(self, name: str) -> Union[str, None]:
        if self.exists():
            return self.stack.output(name)
        elif self.optional:
            log.debug("returning None for %s as stack %s does not exist", name, self.stack.description())
            return None
        else:
            raise StackReferenceException(f"referenced stack {self.ref_name} {self.stack.description()} doesn't exist")

    def __getattr__(self, name: str):
        return self[name]

    def status(self):
        """Delegated to stack"""
        return self.stack.status()

    def exists(self):
        """Delegated to stack"""
        return self.stack.exists()

    def description(self):
        """Delegated to stack"""
        return self.stack.description()

    def outputs(self):
        """Delegated to stack"""
        return self.stack.outputs()

class StackReferenceException(Exception):
    pass

class StackRefs:
    DEFAULTS = {"stack_name": "{{ environment }}-{{ name }}", "optional": False}

    def __init__(self, stack_refs: Dict[str, Dict[str, str]], config: Any):
        self._stacks: Union[Dict[str, OptionalStackReference], None] = None
        self.config = config
        self.refs = stack_refs
        log.debug("defined refs: %s", self.refs)

    def __contains__(self, name: str) -> bool:
        return name in self.stacks()

    def __getitem__(self, name: str) -> OptionalStackReference:
        return self.stack(name)

    def __getattr__(self, name: str):
        if name in self.stacks().keys():
            return self.stacks()[name]
        raise StackReferenceException(f"{name} is not a known stack ref - only know about {', '.join(self.refs.keys())}")

    def exists(self, name: str):
        return self[name].exists()

    def output(self, name: str, output_name: str) -> Union[str, None]:
        """
        Return the output from stack or `None` if stack does not exist and
        stack is optional, otherwise will raise an exception.
        """
        log.info("getting output %s from stack %s", output_name, name)
        stack = self.stack(name)
        return stack.output(output_name)

    def stack(self, name: str) -> OptionalStackReference:
        """
        Returns stack object, or None if stack is optional but is not found
        """
        if name not in self.refs:
            stack_names = sorted(self.refs.keys())
            raise Exception(f"Attempt to access stack {name}, but it's not defined in config.refs - only {', '.join(stack_names)} are defined")

        ref = self.refs[name]
        log.debug("looking up stack %s, ref=%s", name, ref)

        stack = self.stacks()[name]

        log.debug("stack=%s, status=%s, exists=%s", stack, stack.status(), stack.exists())
        if stack.exists():
            log.debug("found stack")
            return stack

        log.debug("stack does not exist, ref=%s", ref)
        if ref['optional']:
            log.debug("stack doesn't exist, but it's optional")
            return stack

        raise Exception(f"Referenced stack {name}, does not exist (stack name={ref['stack_name']}, region={ref.get('region', stack.aws.region)})")

    RESERVED_KEYS = ["environment"]
    def stacks(self) -> Dict[str, OptionalStackReference]:
        if self._stacks is None:
            log.debug("START - building list of stack references")
            # _stacks is dict of {name => OptionalStackReference() for each named stack. This includes
            # stacks that don't exist.
            self._stacks = {}
            for name, cfg in self.refs.items():
                log.debug("processing name=%s: config=%s", name, cfg)
                if name in self.RESERVED_KEYS:
                    log.debug("skipping %s, reserved key", name)
                    continue

                if not cfg:
                    cfg = {}

                if not issubclass(type(cfg), dict):
                    print(f"{name} is not a valid stack reference definition (from {self.refs})")
                    exit(-1)

                # Try building dict of options. This can fail if interpolating incorrect variable or
                #try:
                final_opts = InterpolatedDict({"region": self.config.aws.region, **self.DEFAULTS, **cfg}, {"environment": self.config.environment, "name": name.replace("_", "-"), "region": self.config.aws.region})
                log.debug("Have built final_opts=%s", final_opts)
                #except Exception as ex:
                #    print(f"Unable to process settings for stack reference {name} -> {cfg} (from {self.refs})", ex)
                #    exit(-1)

                try:
                    log.info("final ops: stack reference %s: %s", name, final_opts)
                    opts = StackRefOpts(**final_opts)
                except Exception as ex:
                    log.exception("Invalid configuration for stack.refs '%s' %s: %s", name, self.refs, ex, exc_info=ex)
                    raise

                aws = self.config.aws
                if opts.region != aws.region:
                    log.debug("stack reference is in a different region")
                    aws = copy(aws)
                    aws.region = opts.region

                log.debug("Storing %s as optional stack reference", name)
                stack_ref = OptionalStackReference(name, StackReference(aws=aws, name=opts.stack_name), optional=opts.optional)

                self._stacks[name] = stack_ref
                if stack_ref.exists():
                    try:
                        log.info("stack %s %s has the following outputs: %s", name, stack_ref.description(), stack_ref.outputs())
                    except Exception as ex:
                        log.exception("unable to get outputs for stack %s", name)
                        raise ex
                else:
                    log.info("stack %s %s doesn't exist", name, stack_ref.description())
            log.debug("DONE - building list of stack references")

        return self._stacks



# pylint: disable=missing-docstring

from __future__ import annotations

from pytest import fixture, mark

from . import StackFixtures

from ..stack import StackException
from ..template import RenderedTemplate


class TestStack(StackFixtures):
    @fixture
    def rendered_template(self, cfn_bucket):
        content_bytes = open(self.fixture_path("templates", "minimal.yaml"), "r", encoding="utf-8").read()
        return RenderedTemplate(name="template.yaml", content=content_bytes)

    @mark.skip("moto validation seems to be broken")
    def test_validate(self, stack, rendered_template):
        assert not stack.validate(rendered_template)

    @mark.skip("moto validation seems to be broken")
    def test_new_stack_create_change_set_valid(self, stack, rendered_template):
        res = stack.create_change_set(template=rendered_template, change_set_name="my-changeset-name")
        assert res.available()

        execute_res = stack.execute_change_set(change_set_name="my-changeset-name")
        assert execute_res



import pytest

# pylint: disable=missing-docstring

from . import ConfigFixtures


@pytest.mark.parametrize("config_file", ["includes"], indirect=True)
class TestConfigFile(ConfigFixtures):
    def test_included_mapped(self, config_file):
        assert config_file.includes() == ["includes/first.yaml"]

    def test_load_configs(self, config_file):
        configs = config_file.load_includes()

        assert ["includes/second.yml", "includes/third.yaml", "includes/first.yaml", "main.yaml"] == list(c.filename for c in configs)

    def test_environments(self, config_file):
        assert config_file.environments() == ["dev", "test", "prod"]



from datetime import datetime, timezone
from time import sleep
from botocore.exceptions import ClientError

from rich import box
from rich.table import Table
from rich.live import Live

from . import console

class StackWaiter:
    def __init__(self, stack):
        self.stack = stack
        self.cfn = stack.cfn
        self.seen_events = set()
        self.start_time = datetime.now(timezone.utc)

    class ResourceEvent:
        def __init__(self, *args, **kwargs):
            if kwargs:
                self.logical_id = kwargs["logical_id"]
                self.type = kwargs["type"]
                self.status = "-"
                self.timestamp = None
            elif len(args) > 0:
                assert len(args) == 1
                event = args[0]

                self.logical_id = event["LogicalResourceId"]
                self.type = event["ResourceType"]
                self.event_id = event["EventId"]
                self.status = event["ResourceStatus"]
                self.status_reason = event.get("ResourceStatusReason", None)
                self.timestamp = event["Timestamp"].astimezone()

        def last_seen(self):
            return "%ds ago" % (datetime.now().astimezone() - self.timestamp).seconds if self.timestamp else "-"

    def wait_for_change_set(self, waiter_name, change_set):
        waiter = self.cfn.get_waiter(waiter_name)

        def waiter_callback(response):
            if "Stacks" in response:
                s = response["Stacks"][0]
                # status.update(s["StackStatus"])
                print(s["StackStatus"])
            elif "Error" in response:
                # status.update(response["Error"]["Message"])
                print(response["Error"]["Message"])
                pass

        wrapped_waiter = self.wrap_waiter(waiter, waiter_callback)
        wrapped_waiter.wait(WaiterConfig={"Delay": 5, "MaxAttempts": 720}, StackName=self.stack.name, ChangeSetName=change_set.name)

    def wait_for_stack(self, waiter_name, resources: dict = None):
        """
        Wait for stack change to complete

        :param waiter_name: waiter name. Refer to <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/cloudformation.html#waiters>
        :param resources: used to pre-populate the status update table. It is an (optional) dict with resource logical ID key and resource type as value.
        """
        waiter = self.cfn.get_waiter(waiter_name)

        # If we are applying a change set, we know what resources will be changed, so can
        # pre-populate the table with resources.
        self.resources = {self.stack.name: StackWaiter.ResourceEvent(logical_id=self.stack.name, type="AWS::CloudFormation::Stack")}

        if resources:
            for name, resource_type in resources.items():
                self.resources[name] = StackWaiter.ResourceEvent(logical_id=name, type=resource_type)

        # Wait with a live-updated table showing resources to be changed, and their current
        # state.
        try:
            with Live(self.refresh_table(), refresh_per_second=1, transient=False, console=console) as live:

                def waiter_callback(response):
                    # print(response)
                    live.update(self.refresh_table())
                    if "Stacks" in response:
                        # Update status of stack object (pseudo-resource)
                        stacks = response["Stacks"]
                        if len(stacks) > 0:
                            self.resources[self.stack.name].status = response["Stacks"][0]["StackStatus"]
                    elif "Error" in response:
                        console.log(response["Error"]["Message"])

                waiter = self.wrap_waiter(waiter, waiter_callback)
                waiter.wait(WaiterConfig={"Delay": 2, "MaxAttempts": 1800}, StackName=self.stack.name)

                # Perform a final update so the status table reflects end state
                live.update(self.refresh_table())
        except ClientError as e:
            error_received = e.response["Error"]
            if (error_received["Code"] == "ValidationError") and ("does not exist" in error_received["Message"]):
                return None
            raise (e)

    def refresh_table(self):
        self.process_new_events()
        table = Table("Logical Resource", "Type", "Status", "Last Updated", box=box.SIMPLE)
        for _, resource in sorted(self.resources.items()):
            table.add_row(resource.logical_id, resource.type, resource.status, resource.last_seen())
        return table

    def process_new_events(self):
        paginator = self.cfn.get_paginator("describe_stack_events")
        # Use stack identifier in case stack has been deleted+*
        for events in paginator.paginate(StackName=self.stack.name):
            for e in events["StackEvents"]:
                resource = self.ResourceEvent(e)
                if resource.timestamp >= self.start_time and resource.event_id not in self.seen_events:
                    self.resources[resource.logical_id] = resource
                    self.seen_events.add(resource.event_id)
                    if resource.status_reason:
                        console.log(resource.logical_id + ": " + resource.status_reason)

    def wrap_waiter(self, waiter, callback):
        orig_func = waiter._operation_method

        def wrapper(**kwargs):
            response = orig_func(**kwargs)
            callback(response)
            return response

        waiter._operation_method = wrapper

        return waiter



# pylint: disable=missing-docstring

from os import path, environ

import boto3
from pytest import fixture
from pytest import fixture
from moto import mock_s3, mock_cloudformation, mock_sts


from ..aws_config import AwsSettings
from ..config import Config, ConfigFile
from ..stack import Stack


class Fixtures:
    def fixture_path(self, *dir):
        return path.join(path.dirname(__file__), "fixtures", *dir)

    def fixture_content(self, *dir) -> bytes:
        return open(self.fixture_path(*dir), "rb").read()


class StackFixtures(Fixtures):
    @fixture
    def aws(self):
        """Mocked AWS Credentials for moto."""
        environ["AWS_ACCESS_KEY_ID"] = "testing"
        environ["AWS_SECRET_ACCESS_KEY"] = "testing"
        environ["AWS_SECURITY_TOKEN"] = "testing"
        environ["AWS_SESSION_TOKEN"] = "testing"
        return AwsSettings(region="us-east-1", cfn_bucket="foo")

    @fixture
    def sts(self, aws):
        with mock_sts():
            conn = boto3.client("sts", region_name="us-east-1")
            yield conn

    @fixture
    def s3(self, aws, sts):
        with mock_s3():
            conn = boto3.client("s3", region_name="us-east-1")
            yield conn

    @fixture
    def cloudformation(self, aws, sts):
        with mock_cloudformation():
            conn = boto3.client("cloudformation", region_name="us-east-1")
            yield conn

    @fixture
    def cfn_bucket(self, s3):
        bucket_name = "foo"
        s3.create_bucket(Bucket=bucket_name)
        yield bucket_name

    @fixture
    def stack(self, cloudformation, s3, aws):
        yield Stack(aws=aws, name="test-stack")


class ConfigFixtures(StackFixtures):
    @fixture
    def config(self, request, sts):
        return Config("main", environment="test", config_path=self.fixture_path("config", request.param))

    @fixture
    def config_file(self, request, sts):
        return ConfigFile("main.yaml", self.fixture_path("config", request.param))



"""
  cfn config init -- create a new project
  cfn config add .... -- add config file(s) to existing project
"""
import os
from typing import List
import click

from .init import InitCmd
from .add_template import AddTemplateCmd
from stk import clog


@click.group()
def config():
    """
    Base for performing config-management commands
    """
    pass


@config.command()
@click.argument("directory")
@click.option("--repo", help="Git url for template repository", default="https://github.com/jwoffindin/stk-templates.git")
@click.option("--profile", "-p", help="New config will use this AWS profile")
@click.option("--region", "-r", help="New config will use this region")
@click.option("--bucket", "-b", help="Bucket to use ")
@click.option("--env", "-e", help="Add this environment", multiple=True)
@click.option("--templates", "-t", help="Add template(s)", multiple=True)
def init(directory: str, repo: str, profile: str, region: str, bucket: str, env: List, templates: List) -> None:
    """
    Create a new, config project, as a way to get the user started
    """
    InitCmd(directory).new(
        repo=repo,
        region=region,
        profile=profile,
        bucket=bucket,
        environments=env
    )

    for template_list in templates:
        for tpl in template_list.split(","):
            # try:
            AddTemplateCmd(repo, config_dir=directory).add(tpl)
            # except Exception as ex:
            #    clog(f"unable to add {tpl}: {ex}")


@config.command()
@click.argument("template")
@click.option("--repo", help="Path to template project")
@click.option("--inline", is_flag=True, default=False, help="Create a local copy of template")
@click.option("--template-dir", default=os.environ.get('TEMPLATE_PATH', 'templates'), help="Where --copy-template should store the template")

def add(template: str, repo: str, inline: bool, template_dir: str) -> None:
    """
    Add configuration for template from remote repository
    """
    AddTemplateCmd(repo).add(template, inline=inline, local_template_dir=template_dir)



import os
import random
import sys

from dataclasses import dataclass
from typing import List

import boto3
import inquirer
import yaml
import botocore

from git.repo import Repo

from .. import clog


class InitCmd:
    """
    Interactive process taking the user through the setup of a new
    config project.

    Creates a new directory, writes configuration files, creates new git repo
    and adds files.
    """

    def __init__(self, directory) -> None:
        self.config = {}
        self.directory = directory
        self.repo: Repo

    def new(self, **kwargs):
        """
        Entrypoint
        """
        self._build_config(**kwargs)

        self._create_config_dir()
        self._git_init()
        self._write_config_files()
        self._commit_config_files()

    def _build_config(self, repo: str, profile: str = "", region: str = "", bucket: str = "", environments: List = []):
        """
        Sets self.config to be hash of files and their content that are to be written to
        new config project
        """
        if os.path.exists(self.directory):
            print(f"{self.directory} already exists")
            sys.exit(-2)

        if not environments:
            environments = self._gather_environments()

        aws = self._gather_aws_settings(
            profile=profile,
            region=region,
            bucket_name=bucket
        )

        common_config = {
            **aws,
            "environments": {x: None for x in environments},
            "core": {"environments": environments},
        }

        common_config["template"] = {
            "root": "/",
            "version": "main",
            "repo": repo,
        }

        # This is really only useful for template development
        # Probably put this in a toggle, or just document it
        # if "dev" in environments:
        #     envs = common_config["environments"]
        #     if envs["dev"] is None:
        #         envs["dev"] = {}
        #
        #     envs["dev"]["template"] = {
        #         "version": None,
        #         "repo": None,
        #         "root": '{{ environ["TEMPLATE_PATH"] }}',
        #     }

        yaml.SafeDumper.add_representer(
            type(None),
            lambda x, value: x.represent_scalar("tag:yaml.org,2002:null", "")
        )

        self.config = {
            "includes/common.yml": yaml.safe_dump(
                common_config,
                default_flow_style=False,
                explicit_start=True, sort_keys=False
            ),
            ".gitignore": "*.log\n.template-cache\n"
        }

    def _create_config_dir(self):
        clog(f"creating new config project {self.directory}")
        os.mkdir(self.directory)

    def _git_init(self):
        clog("initializing new git repository")
        self.repo = Repo.init(self.directory)

    def _write_config_files(self):
        clog("writing configuration files")
        for filename, content in self.config.items():
            full_path = os.path.join(self.directory, filename)
            dir_name = os.path.dirname(full_path)
            if not os.path.exists(dir_name):
                os.mkdir(dir_name)

            with open(full_path, "x", encoding="utf-8") as fh:
                fh.write(content)

    def _commit_config_files(self):
        clog("committing new files")
        self.repo.index.add(list(self.config.keys()))
        self.repo.index.commit(
            "Initial configuration added by cfn config init"
        )

    @dataclass
    class GatherAwsSettings:
        """
        Build settings for AWS - optionally creating an S3 bucket if required
        """
        region: str
        profile: str
        bucket_name: str

        def gather(self):
            """
            Handle process of collecting AWS settings - profile, region and CFN bucket
            """
            questions = []

            self._gather_profile(questions)
            self._gather_region(questions)

            aws_settings = {
                "profile": self.profile,
                "region": self.region,
            }

            answers = inquirer.prompt(questions)
            if answers:
              aws_settings = {**aws_settings, **answers}

            self._gather_bucket(**aws_settings)

            aws_settings["cfn_bucket"] = self.bucket_name

            return aws_settings

        def _gather_profile(self, questions: List):
            session = boto3.session.Session()
            if not self.profile:
                profiles = session.available_profiles
                if not profiles:
                    print(
                        "Your aws configuration does not set any profiles. Run `aws configure`")
                    sys.exit(-1)

                # Don't bother if user only has one profile configured
                if self.profile is None:
                    if len(profiles) == 1:
                        self.profile = profiles[0]

                if not self.profile:
                    questions.append(
                        inquirer.List(
                            "profile",
                            message="Select AWS Profile to use",
                            choices=profiles,
                        )
                    )

        def _gather_region(self, questions: List):
            session = boto3.session.Session()
            if self.region is None:
                regions = sorted(session.get_available_regions("ec2"))
                questions.append(
                    inquirer.List(
                        "region",
                        message="What AWS region to you want to deploy to?",
                        choices=regions,
                        default="us-west-2"
                    )
                )

        def _gather_bucket(self, region: str, profile: str):
            # Now that we have profile/region - we can figure out what bucket to use
            if not self.bucket_name:
                session = boto3.session.Session(
                    profile_name=profile, region_name=region)

                try:
                    s3_client = session.client("s3")
                    create_option = "NONE - create a new bucket"

                    buckets = [create_option]
                    for bucket_name in s3_client.list_buckets()["Buckets"]:
                        name = bucket_name['Name']
                        location = s3_client.get_bucket_location(Bucket=name)
                        if (region == 'us-east-1' and location['LocationConstraint'] == None) or location['LocationConstraint'] == region:
                            buckets.append(name)

                    bucket_answer = inquirer.prompt([
                        inquirer.List(
                            "bucket", message="Select bucket to use for cloudformation templates", choices=buckets)
                    ])

                    if not bucket_answer:
                        print("aborting")
                        sys.exit(-1)

                    if bucket_answer["bucket"] != create_option:
                        self.bucket_name = bucket_answer['bucket']
                        return

                    self.bucket_name = f"cf-templates-{region}-{'%015x' % random.getrandbits(60)}"

                    clog(f"creating s3 bucket {self.bucket_name}")

                    if region == 'us-east-1':
                      s3_client.create_bucket(Bucket=self.bucket_name)
                    else:
                      s3_client.create_bucket(Bucket=self.bucket_name, CreateBucketConfiguration={'LocationConstraint': region})

                    s3_client.put_public_access_block(
                        Bucket=self.bucket_name,
                        PublicAccessBlockConfiguration={
                            'BlockPublicAcls': True,
                            'IgnorePublicAcls': True,
                            'BlockPublicPolicy': True,
                            'RestrictPublicBuckets': True
                        },
                    )

                except botocore.exceptions.NoCredentialsError as ex:
                    print(
                        f"Unable to retrieve list of buckets using {profile}: {ex}")
                    sys.exit(-1)

    def _gather_aws_settings(self, **kwargs) -> dict:
        """
        Gather information about AWS deployment
        """
        settings = self.GatherAwsSettings(**kwargs).gather()

        return {"aws": settings}

    def _gather_environments(self) -> List:
        """
        Allow user to specify what environment's to deploy
        """
        answers = inquirer.prompt(
            [
                inquirer.Checkbox(
                    "environments",
                    message="What environments do you want to deploy to?",
                    choices=["dev", "test", "staging", "prod"],
                    default=["dev"],
                )
            ]
        )

        if answers:
            envs = answers["environments"]
            if envs:
                return envs

        return ["dev, test", "prod"]


if __name__ == "__main__":
    InitCmd("tmp").new()



# pylint: disable=missing-docstring

import pytest

from pytest import fixture, mark
from click.testing import CliRunner

from . import StackFixtures
from ..cli import stk
from ..template import Template


class CliFixtures(StackFixtures):
    @fixture
    def args(self, request, s3, cloudformation, cfn_bucket):
        config_path = self.fixture_path("cli", "config")
        template_path = self.fixture_path("cli", "templates")
        command = request.param.pop(0)
        return [command, *request.param, "--config-path", config_path, "--template-path", template_path]

    @fixture
    def cli(self, args):
        return CliRunner().invoke(stk, args)


class TestCliShowTemplate(CliFixtures, StackFixtures):
    @mark.parametrize("args", [["show-template", "basic", "test"]], indirect=True)
    def test_show_basic_template(self, cli):
        assert "This is a sample template for test environment" in cli.output
        assert 'TopicName: "a_topic"' in cli.output
        assert cli.exit_code == 0

    @mark.parametrize("args", [["show-template", "different-template-name", "test"]], indirect=True)
    def test_show_failed_template(self, cli):
        assert 'TopicName: "another_topic"' in cli.output
        assert cli.exit_code == 0


class TestCliValidateTemplate(CliFixtures):
    @mark.parametrize("args", [["validate", "basic", "test"]], indirect=True)
    @mark.skip("I think this should be a valid test")
    def test_validate_basic_template(self, cli):
        assert "Template is ok" in cli.output
        assert cli.exit_code == 0

    @mark.parametrize("args", [["validate", "invalid", "test"]], indirect=True)
    def test_invalid_yaml_template(self, args):
        cli = CliRunner().invoke(stk, args, catch_exceptions=False)

        assert "Template is NOT ok" in cli.output
        assert cli.exit_code == -1


# class TestCliCreate(CliFixtures):
#     @mark.parametrize('args', [['create', 'basic', 'test']], indirect=True)
#     def test_create_stack(self, cli):
#         assert 'Creating stack test-basic' in cli.output
#         assert 'Stack created successfully' in cli.output
#         assert cli.exit_code == 0


class TestCliCreateChangeSet(CliFixtures):
    @mark.parametrize("args", [["create-change-set", "basic", "test", "my-change-set"]], indirect=True)
    def test_create_change_set(self, cli):
        assert "Creating change set my-change-set for test-basic" in cli.output
        assert "Change set created" in cli.output
        assert cli.exit_code == 0



from __future__ import annotations

import boto3
import botocore

from dataclasses import dataclass

from .aws_config import AwsSettings


class Uploadable:
    def body(self) -> bytes:
        pass

    def key(self) -> str:
        pass


@dataclass
class CfnBucketObject:
    bucket: CfnBucket
    key: str

    def as_http(self) -> str:
        """
        Return http url for object
        """
        s3_hostname = f"{self.bucket.bucket_name}.s3.{self.bucket.region}.amazonaws.com"
        return "/".join(["https:/", s3_hostname, self.key])

    def as_s3(self) -> str:
        """
        Return s3:// url for object
        """
        return "/".join(["s3:/", self.bucket.bucket_name, self.key])


class CfnBucket:
    FORCE_OVERWRITE = False

    def __init__(self, config: AwsSettings):
        self.bucket_name = config.cfn_bucket
        self.region = config.region
        self.s3 = config.client("s3")

    # Upload content object to S3 bucket
    def upload(self, object: Uploadable, overwrite: bool = False):
        s3 = self.s3

        if self.FORCE_OVERWRITE:
            s3.put_object(Bucket=self.bucket_name, Key=object.key(), Body=object.body(), ServerSideEncryption="AES256")
        else:
            try:
                s3.head_object(Bucket=self.bucket_name, Key=object.key())
                # print(f"Key {object.key()} already exists in {self.bucket_name}, not uploading")
            except botocore.exceptions.ClientError as ex:
                if ex.response["ResponseMetadata"]["HTTPStatusCode"] == 404:
                    s3.put_object(Bucket=self.bucket_name, Key=object.key(), Body=object.body(), ServerSideEncryption="AES256")
                else:
                    raise
        return CfnBucketObject(self, object.key())



"""
"""
import logging
import logging.handlers
import yaml

from os import environ
from rich.console import Console

VERSION = "0.15.0"

# Create application logger (for when things go wrong)
log = logging.getLogger('cfn')
log.setLevel(environ.get('CFN_LOG_LEVEL', environ.get("LOG_LEVEL", "WARN")).upper())

if environ.get('LOG_FILE'):
    logging.basicConfig(filename=environ['LOG_FILE'], filemode="a")
else:
    handler = logging.handlers.SysLogHandler(address = '/dev/log')
    formatter = logging.Formatter('cfn: %(message)s')
    log.addHandler(handler)


# Console logger is for displaying updates to user - normal
# events.
console = Console(emoji=False, log_path=False, stderr=True)
clog = console.log

# This makes yaml.dump() output `foo: ` rather than `foo: null`
#
yaml.SafeDumper.add_representer(
    type(None),
    lambda x, value: x.represent_scalar("tag:yaml.org,2002:null", "")
)


class ConfigException(Exception):
    """
    Needs a new home - supposed to be base class for any configuration
    based errors.
    """
    pass



"""
Abstracts access templates stored in either a remote git
repository, or local filesystem.
"""
from __future__ import annotations

import os
import stat
import types
from typing import Union, cast
import urllib

from os import path, walk
from pathlib import Path
from dataclasses import dataclass

import giturlparse

from git.repo import Repo
from git import Blob, Tree
from gitdb.exc import BadName

from . import log


class GenericProvider:
    def template(self) -> bytes:
        tpl_path = Path(self.name)
        if not tpl_path.suffix:
            tpl_path = tpl_path.with_suffix(".yaml")
        return self.content(tpl_path)

    def content(self, _: str) -> bytes:
        pass

    def is_file(self, *_) -> bool:
        pass

    def is_dir(self, *_) -> bool:
        pass

    def find(self, dir: str, ignore: function):
        pass


@dataclass
class FilesystemProvider(GenericProvider):
    name: str
    root: str

    def __post_init__(self):
        # Provider needs to work with absolute paths; $CWD is changed during
        # various processing changes, so specifying relative template path
        # breaks helpers etc.
        self.root = path.abspath(self.root)

        # Check the path actually exists
        if not Path(self.root).is_dir:
            raise Exception(f"{self.root} does not appear to be a directory")

    def head(self):
        return None

    def content(self, file_path: str) -> bytes:
        return open(path.join(self.root, file_path), "rb").read()

    def is_file(self, *p) -> bool:
        return path.isfile(path.join(self.root, *p))

    def is_tree(self, *p) -> bool:
        return path.isdir(path.join(self.root, *p))

    def find(self, dir, ignore: types.FunctionType = None):  # type: ignore
        start_dir = path.abspath(path.join(self.root, dir))

        if not path.exists(start_dir):
            raise Exception(f"{dir} does not exist in {self.root}")

        if self.is_tree(start_dir):
            log.info("adding directory tree under %s to zip", dir)
            for p, _, files in walk(start_dir):
                for f in files:
                    file_path = path.join(p, f)
                    if ignore and ignore(file_path):
                        log.info("Skipping %s due to ignore rule", file_path)
                        continue
                    # strip the directory prefix. E.g.
                    #  'functions/<function-name>/foo.txt' -> foo.txt
                    relative_path = file_path[len(start_dir) + 1 :]

                    file_st = os.lstat(file_path)
                    if stat.S_ISREG(file_st.st_mode):
                        yield (relative_path, "file", self.content(file_path))
                    elif stat.S_ISLNK(file_st.st_mode):
                        target = os.readlink(file_path)
                        yield (relative_path, "symlink", target)
                    else:
                        raise Exception("Unsupported filesystem object at " + file_path)
        elif self.is_file(start_dir):

            relative_path = path.basename(start_dir)

            log.info("adding single file %s (from %s) to zip", relative_path, dir)

            if ignore and ignore(start_dir):
                log.warning("Adding a single file to zip, but it's in the ignore list")

            file_st = os.lstat(start_dir)
            if stat.S_ISREG(file_st.st_mode):
                yield (relative_path, "file", self.content(start_dir))
            elif stat.S_ISLNK(file_st.st_mode):
                target = os.readlink(start_dir)
                yield (relative_path, "symlink", target)
            else:
                raise Exception("Unsupported filesystem object at " + start_dir)
        else:
            raise Exception("Unsupported filesystem object at " + start_dir)

    def __str__(self) -> str:
        return self.name


@dataclass
class GitProvider(GenericProvider):
    name: str
    git_url: str  # TODO rename to root
    git_ref: str
    root: str = ""

    #
    repo: Repo = None

    def __post_init__(self):
        if self.root.endswith("/"):
            self.root = self.root[:-1]

        log.debug(f"GitProvider(name={self.name}, url={self.git_url}, root={self.root})")

        if not self.git_url:
            raise Exception("template.git_url is not set")

        if self.git_url.startswith(".") or self.git_url.startswith("/"):
            # Local repository (e.g. ../templates)
            self.repo = Repo(self.git_url)
        else:
            url = urllib.parse.urlparse(self.git_url)
            log.debug(url)
            if url.scheme == "file":  # "file" in url.protocols:
                log.info("Local git repository at %s", url.path)
                self.repo = Repo(url.path)
            else:
                # Remote repository
                url = giturlparse.parse(self.git_url)
                cache_dir = self._cache_path(url)
                if path.exists(cache_dir):
                    log.info("using existing cached version %s", cache_dir)
                    self.repo = Repo(cache_dir)
                    log.info("git fetch remote 'origin'")
                    self.repo.remotes["origin"].fetch(refspec="+refs/heads/*:refs/heads/*")
                else:
                    log.info("don't have a cached copy of %s in %s", self.git_url, cache_dir)
                    os.makedirs(cache_dir, mode=0o700, exist_ok=True)
                    log.info("cloning from %s -> %s", self.git_url, cache_dir)
                    self.repo = Repo.clone_from(self.git_url, cache_dir, bare=True)

        log.info("getting commit %s from %s", self.git_ref, self.repo)

        try:
            self.commit = self.repo.commit(self.git_ref)
        except BadName as exc:
            log.info("can't resolve %s, trying with origin/prefix", self.git_ref)
            try:
                self.commit = self.repo.commit("origin/" + self.git_ref)
            except BadName:
                raise exc from exc # throw the original error

        log.info("have commit %s", self.commit.hexsha)

    def content(self, *p) -> bytes:
        file_path = path.join(self.root, *p)
        try:
            log.info(f"getting content for {file_path}")
            return self.commit.tree[file_path].data_stream.read()
        except KeyError as ex:
            log.exception(f"Git object (git={self.git_url} file_path={file_path})@{self.git_ref}: does not exist", exc_info=ex)
            raise

    def head(self):
        return self.commit

    def is_file(self, *p) -> bool:
        file_path = path.join(self.root, *p)
        try:
            obj = self.commit.tree[file_path]
            if obj.type == "blob":
                return True
            return False
        except KeyError:
            return False

    def is_tree(self, *p) -> bool:
        dir_path = path.join(self.root, *p)
        try:
            obj = self.commit.tree[dir_path]
            if obj.type == "tree":
                return True
            return False
        except KeyError:
            return False

    def find(self, dir, ignore: types.FunctionType = None):  # type: ignore
        if dir.endswith("/"):
            dir = dir[:-1]

        tree_path = path.join(self.root, dir)
        tree = self.commit.tree[tree_path]
        if tree.type == "tree":
            log.info("adding directory tree under %s to zip", tree_path)
            for item in tree.traverse():
                item = cast(Union['Tree', 'Blob'], item)
                item_path = item.path[len(tree_path) + 1 :]
                if ignore and ignore(item_path):
                    continue

                if item.type == "blob":
                    if item.mode & item.link_mode == item.link_mode:
                        type = "symlink"
                    else:
                        type = "file"
                    yield (item_path, type, item.data_stream.read())
        elif tree.type == "blob":
            log.info("adding single file %s to zip", tree_path)
            if tree.mode & tree.link_mode == tree.link_mode:
                type = "symlink"
            else:
                type = "file"
            yield (tree_path, type, tree.data_stream.read())
        else:
            raise Exception("Unsupported git object at " + tree_path)

    def _cache_path(self, url: giturlparse.parser.Parsed) -> str:
        cache_dir = os.environ.get("TEMPLATE_CACHE", ".template-cache")
        log.debug(f"setting cache path from url {url}")
        return path.join(cache_dir, url.host, url.owner, *(url.groups), url.repo)  # type: ignore


def provider(source):
    log.info(f"provider = {source}")
    if source.version:
        root = source.root or "/"
        return GitProvider(name=source.name, git_url=source.repo, root=root, git_ref=source.version)
    else:
        return FilesystemProvider(name=source.name, root=source.root)



#!/usr/bin/env python3 -m stk.cli

from __future__ import annotations

import functools
import json

from typing import List
from os import environ

import click
import boto3
import yaml
from click_aliases import ClickAliasedGroup
from rich.table import Table
from rich.prompt import Confirm
from rich.padding import Padding

from . import VERSION, console
from .config import Config
from .config_cmd import cli
from .stack_delegated_command import StackDelegatedCommand
from .template import TemplateWithConfig
from .template_command import TemplateCommand
from .util import parse_overrides

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


def common_stack_params(func):
    """Decorator for commands that need same stack/environment parameters"""

    @click.argument("name")
    @click.argument("environment")
    @click.option("--config-path", default=environ.get("CONFIG_PATH", "."), help="Path to config project")
    @click.option("--template-path", default=environ.get("TEMPLATE_PATH", "."), help="Path to templates")
    @click.option("--var", help="override configuration variable", multiple=True)
    @click.option("--param", help="override CFN parameter", multiple=True)
    @click.option("--overrides", help="override generic config")
    @click.option("--outputs-format", type=click.Choice(['table', 'json', 'yaml'], case_sensitive=False), default="table", help="Force conversion of resulting template (json, yaml or table)")
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


@click.group(context_settings=CONTEXT_SETTINGS, cls=ClickAliasedGroup)
@click.version_option(version=VERSION)
def stk():
    """
    Primary command object for CLI
    """
    log_level = environ.get("LOG_LEVEL", None)
    if log_level:
        boto3.set_stream_logger("boto3", level=log_level)
        boto3.set_stream_logger("botocore", level=log_level)
        boto3.set_stream_logger("boto3.resources", level=log_level)


# Register config subcommands
stk.add_command(cli.config)


@stk.command()
@common_stack_params
def validate(**kwargs):
    """Validate provided template"""
    sc = TemplateCommand(**kwargs)
    sc.validate()


@stk.command()
@common_stack_params
@click.option("--yes", is_flag=True, show_default=True, default=False, help="Automatically approve changeset")
def create(yes: bool, **kwargs):
    """
    Create a new stack
    """
    sc = TemplateCommand(**kwargs)
    if sc.exists():
        console.log(f"Stack {sc.stack_name} already exists", style="red")
        exit(-1)

    console.log("Creating stack", sc.stack_name, style="blue")

    with console.status("Validating template"):
        sc.validate()

    change_set = sc.create_change_set()
    console.log("Change set created")

    console.print(change_set.summary())

    if not change_set.available():
        console.log(":x: Change set could not be generated",
                    emoji=True, style="red")
        exit(-2)

    if yes or Confirm.ask(f"Create stack {sc.stack_name} ?"):
        console.rule()
        console.log("Applying change set")
        change_set.execute()
        if sc.wait("stack_create_complete", change_set.resources()):
            console.log("Stack created successfully", style="green")
            if sc.outputs():
                sc.show_outputs()
        else:
            console.log("Stack create failed", style="red")
            exit(-2)
    else:
        console.log("Cleaning up")
        sc.delete()


@stk.command()
@common_stack_params
@click.option("--yes", is_flag=True, show_default=True, default=False, help="Automatically approve changeset")
def update(yes: bool, **kwargs):
    """
    Update an existing stack via changeset
    """
    sc = TemplateCommand(**kwargs)

    if not sc.exists():
        console.log(f"Stack {sc.name} does not exist in {sc.config.aws.region}")
        exit(-1)

    sc.validate()

    console.log("Diff:\n", sc.diff(sc.template))

    change_set = sc.create_change_set()

    console.log("Change set:\n", change_set.summary())

    if not change_set.available():
        if change_set.is_empty_changeset():
            console.log(":poop: No changes to be applied",
                        emoji=True, style="blue")
            exit(-9)
        else:
            console.log(":x: Change set could not be generated",
                        emoji=True, style="red")
            exit(-2)

    if yes or Confirm.ask(f"Update stack {sc.stack_name} ?"):
        console.rule()
        console.log("Applying change set")

        change_set.execute()
        if sc.wait("stack_update_complete", change_set.resources()):
            console.log("Stack updated successfully", style="green")
            if sc.outputs():
                sc.show_outputs()
        else:
            console.log("Stack update failed", style="red")
            exit(-2)


@stk.command(aliases=['deploy'])
@common_stack_params
@click.option("--yes", is_flag=True, show_default=True, default=False, help="Automatically approve changeset")
@click.pass_context
def upsert(context, yes: bool, **kwargs):
    """
    Create or update stack
    """
    stack = TemplateCommand(**kwargs)

    if stack.exists():
        context.forward(update, yes=yes)
    else:
        context.forward(create, yes=yes)


@stk.command()
@common_stack_params
@click.argument("change_set_name")
def create_change_set(change_set_name: str, **kwargs):
    """
    Create named changeset
    """
    sc = TemplateCommand(**kwargs)

    console.log(f"Creating change set {change_set_name} for {sc.stack_name}")

    # Fail fast
    sc.validate()

    if sc.exists():
        # Only diff if the stack exists
        console.log("Diff:\n", sc.diff(sc.template))

    change_set = sc.create_change_set(change_set_name=change_set_name)

    console.print(Padding(change_set.summary(), (0, 10)))

    if change_set.available():
        console.log(":+1: Change set created", emoji=True)
    else:
        console.log(":x: Change set was not successful", emoji=True)


@stk.command()
@common_stack_params
@click.argument("change_set_name")
def execute_change_set(change_set_name: str, **kwargs):
    sc = StackDelegatedCommand(**kwargs)

    console.log(f"Executing change set {change_set_name} for {sc.stack_name}")
    try:
        StackDelegatedCommand(
            **kwargs).execute_change_set(change_set_name=change_set_name)
        console.log(":+1: Change set complete", emoji=True, style="green")
        if sc.outputs():
            sc.show_outputs()
    except sc.cfn.exceptions.ChangeSetNotFoundException:
        console.log(":x: Change set does not exist", emoji=True, style="red")


@stk.command()
@common_stack_params
@click.argument("change_set_name")
@click.option("--yes", is_flag=True, show_default=True, default=False, help="Automatically approve changeset")
def delete_change_set(change_set_name: str, yes: bool, **kwargs): # pylint: disable=unused-argument
    """
    Deletes named change set.
    """
    sc = StackDelegatedCommand(**kwargs)
    console.log(f"Deleting change set {change_set_name} for {sc.stack_name}")
    try:
        sc.delete_change_set(change_set_name=change_set_name)
        console.log("Change set deleted")
    except sc.cfn.exceptions.ChangeSetNotFoundException:
        console.log(":x: Change set does not exist", style="red", emoji=True)


@stk.command()
@common_stack_params
@click.option("--yes", is_flag=True, show_default=True, default=False, help="Automatically approve changeset")
def delete(yes: bool, **kwargs):
    sc = StackDelegatedCommand(**kwargs)
    if not sc.exists():
        console.log(f"Stack {sc.stack_name} does not exist in {sc.config.aws.region}", style="red")
        exit(-1)

    if not (yes or Confirm.ask(f"Delete stack {sc.stack_name} ?")):
        console.log("Aborting")
        exit(-2)

    console.rule()
    console.log(f"Destroying stack {sc.stack_name}")
    sc.delete()
    console.log("Stack deleted")


@stk.command()
@common_stack_params
@click.option("--format", "format_", default="yaml", help="Force conversion of resulting template (json or yaml)")
@click.option("--output-file", default="", help="Write resulting template to file")
def show_template(name: str, environment: str, config_path: str, template_path: str, format_: str, output_file: str, outputs_format: str, var: List, param: List, overrides: str):
    config_overrides = parse_overrides(var, param, overrides)
    config = Config(
        name=name,
        environment=environment,
        config_path=config_path,
        template_path=template_path,
        overrides=config_overrides,
    )
    template = TemplateWithConfig(
        provider=config.template_source.provider(), config=config)

    result = template.render()
    if result.error:
        console.log(
            f":x: Template is NOT ok - {result.error}", emoji=True, style="red")
        console.print(str(result))
        exit(-1)

    result = str(result)

    if format_ == "yaml":
        pass
    elif format_ == "json":
        result = json.dumps(yaml.safe_load(result), indent=4, sort_keys=True)
    else:
        raise Exception("Unknown output format")

    if output_file:
        with open(output_file, "w", encoding="utf-8") as file:
            file.write(result)
    else:
        console.print(result)


@stk.command()
@common_stack_params
def diff(**kwargs):
    sc = TemplateCommand(**kwargs)
    if not sc.exists():
        console.log(f"Stack {sc.name} does not exist in {sc.config.aws.region}", style="red")
        return

    d = sc.diff(sc.template)
    if d:
        console.log("Generated diff")
        console.print(d)
        # console.log(d)
    else:
        console.log("There are no changes (templates are identical)")


@stk.command()
@common_stack_params
def outputs(**kwargs):
    """Command to display stack outputs"""
    sc = StackDelegatedCommand(**kwargs)

    console.log(f"Retrieving outputs for {sc.name}")

    if not sc.exists():
        console.log(f"Stack {sc.name} does not exist in {sc.config.aws.region}", style="red")
        return

    sc.show_outputs()


@stk.command()
@common_stack_params
def show_config(name: str, environment: str, config_path: str, template_path: str, outputs_format: str, var: List, param: List, overrides: str):
    config = Config(name=name, environment=environment, config_path=config_path, overrides=parse_overrides(var, param, overrides))

    template = config.template_source
    template_table = Table("Property", "Value", title="Template Source")
    template_table.add_row("Template Name", template.name)
    template_table.add_row("Version", template.version)
    template_table.add_row("Source", template.repo)

    params = config.params
    params_table = Table("Parameter", "Value", "Type", title="Parameters")
    for k in sorted(params.keys()):
        v = params[k]
        params_table.add_row(k, str(v), type(v).__name__)

    cfg_vars = config.vars
    vars_table = Table("Variable", "Value", "Type", title="Variables")
    for k in sorted(cfg_vars.keys()):
        v = cfg_vars[k]
        vars_table.add_row(k, str(v), type(v).__name__)

    tags = config.tags
    tags_table = Table("Tag", "Value")
    for k in sorted(tags.keys()):
        v = tags[k]
        tags_table.add_row(k, str(v))

    console.print(template_table, "\n")
    console.print(params_table, "\n")
    console.print(vars_table, "\n")
    console.print(tags_table, "\n")


if __name__ == "__main__":
    stk()



from __future__ import annotations

from typing import List
from random import choice
from string import ascii_letters
from botocore.exceptions import ClientError, WaiterError
from difflib import unified_diff

from . import log
from .stack_waiter import StackWaiter
from .template import RenderedTemplate
from .basic_stack import BasicStack
from .change_set import ChangeSet


class StackException(Exception):
    def __init__(self, stack: Stack, message: str, response=None):
        super().__init__(message)
        self.stack = stack
        self.response = response


class Stack(BasicStack):
    DIFF_CONTEXT_LINES = 5

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def validate(self, template: RenderedTemplate):
        """Validate template"""
        log.warning("Validating template")
        try:
            self.cfn.validate_template(TemplateURL=self.bucket.upload(template).as_http())
        except ClientError as ex:
            log.warning(ex.response)
            if ex.response["Error"]["Code"] == "ValidationError":
                return ex
            raise

    def create_change_set(self, template: RenderedTemplate, tags: List(str) = [], change_set_name: str = None, params: dict = {}) -> ChangeSet:
        """
        Creates a ChangeSet object for given template.

        :param change_set_name: Change set name. Short random string will be generated if none provided

        """
        if not change_set_name:
            change_set_name = "".join((choice(ascii_letters) for _ in range(12)))

        status = self.status()
        change_set_type = "UPDATE" if status and status != "REVIEW_IN_PROGRESS" else "CREATE"

        return ChangeSet(self, change_set_name).create(template=template, change_set_type=change_set_type, tags=tags, params=params)

    def execute_change_set(self, change_set_name: str) -> bool:
        """
        Execute named change set.

        Returns truthy value if successful
        """

        waiter_name = "stack_update_complete" if self.status() != "REVIEW_IN_PROGRESS" else "stack_create_complete"

        change_set = ChangeSet(self, change_set_name)
        change_set.execute()

        try:
            StackWaiter(self).wait_for_stack(waiter_name, change_set.resources())
            return True
        except WaiterError as ex:
            print("Change set could not be applied:", ex)

    def delete_change_set(self, change_set_name: str):
        ChangeSet(self, change_set_name).delete()

    def delete(self):
        # Get list of resources before we start deleting them
        resources = self.resources()
        self.cfn.delete_stack(StackName=self.name)
        StackWaiter(self).wait_for_stack("stack_delete_complete", resources=resources)

    def diff(self, template: RenderedTemplate) -> str:
        """
        Return text-based diff between currently deployed stack template and the proposed
        (local) template.
        """
        # Compare currently deployed stack template...
        current_template = self.cfn.get_template(StackName=self.name, TemplateStage="Original")
        from_lines = current_template["TemplateBody"].splitlines()

        # against the one just compiled
        to_lines = template.content.splitlines()

        changes = list(unified_diff(from_lines, to_lines, fromfile="Deployed", tofile="Local", n=self.DIFF_CONTEXT_LINES, lineterm=""))

        if not len(changes):
            return "[yellow]There are no changes (template is identical)[/yellow]"

        diff = []
        for change in changes:
            if change.startswith("@@"):
                diff.append(f"[bold]{change}[/bold]")
            elif change.startswith("-"):
                diff.append(f"[red]{change}[/red]")
            elif change.startswith("+"):
                diff.append(f"[green]{change}[/green]")
            else:
                diff.append(change)

        return "\n".join(diff)

    def resources(self) -> dict:
        ret_val = {}

        paginator = self.cfn.get_paginator("list_stack_resources")
        for resources in paginator.paginate(StackName=self.name):
            for resource in resources["StackResourceSummaries"]:
                ret_val[resource["LogicalResourceId"]] = resource["ResourceType"]

        return ret_val

    def wait(self, waiter_name, resources):
        try:
            StackWaiter(self).wait_for_stack(waiter_name, resources)
            return True
        except WaiterError as ex:
            print("Change could not be applied:", ex)
